import math

import pytest

from ase import Atoms
from ase.build import molecule
from ase.utils.ff import VdW, get_vdw_potential_value


@pytest.fixture()
def atoms() -> Atoms:
    mol = molecule("H2")
    mol.center(5.0)
    return mol


def test_utils_ff_vdw(atoms: Atoms) -> None:
    """Check the equivalent van der Waals energies."""
    e1, e2, s1, s2, scale = 0.646, 1.557, 0.777, 0.724, 1.2236

    vdw_a = VdW(
        0, 1, epsiloni=e1, epsilonj=e2, sigmai=s1, sigmaj=s2, scale=scale
    )

    vdw_p = VdW(
        0, 1, epsilonij=math.sqrt(e1 * e2), sigmaij=(s1 + s2) / 2.0, scale=scale
    )

    e_ref = get_vdw_potential_value(atoms, vdw_a)
    assert get_vdw_potential_value(atoms, vdw_p) == pytest.approx(e_ref)



import pytest

from ase.build import molecule
from ase.mep import NEB, idpp_interpolate


# I think idpp uses an optimizer;
# at the very least how an optimizer is called during this test
@pytest.mark.optimize()
def test_idpp(testdir):
    initial = molecule('C2H6')
    final = initial.copy()
    final.positions[2:5] = initial.positions[[3, 4, 2]]

    images = [initial]
    for _ in range(5):
        images.append(initial.copy())
    images.append(final)

    neb = NEB(images)
    d0 = images[3].get_distance(2, 3)
    neb.interpolate()
    d1 = images[3].get_distance(2, 3)
    idpp_interpolate(neb, fmax=0.005)
    d2 = images[3].get_distance(2, 3)
    print(d0, d1, d2)
    assert abs(d2 - 1.74) < 0.01



"""ASE-interface to Octopus.

Ask Hjorth Larsen <asklarsen@gmail.com>
Carlos de Armas

http://tddft.org/programs/octopus/
"""

import numpy as np

from ase.calculators.genericfileio import (
    BaseProfile,
    CalculatorTemplate,
    GenericFileIOCalculator,
)
from ase.io.octopus.input import generate_input, process_special_kwargs
from ase.io.octopus.output import read_eigenvalues_file, read_static_info


class OctopusIOError(IOError):
    pass


class OctopusProfile(BaseProfile):
    def get_calculator_command(self, inputfile):
        return []

    def version(self):
        import re
        from subprocess import check_output
        txt = check_output([*self._split_command, '--version'],
                           encoding='ascii')
        match = re.match(r'octopus\s*(.+)', txt)
        # With MPI it prints the line for each rank, but we just match
        # the first line.
        return match.group(1)


class OctopusTemplate(CalculatorTemplate):
    _label = 'octopus'

    def __init__(self):
        super().__init__(
            'octopus',
            implemented_properties=['energy', 'forces', 'dipole', 'stress'],
        )
        self.outputname = f'{self._label}.out'
        self.errorname = f'{self._label}.err'

    def read_results(self, directory):
        """Read octopus output files and extract data."""
        results = {}
        with open(directory / 'static/info') as fd:
            results.update(read_static_info(fd))

        # If the eigenvalues file exists, we get the eigs/occs from that one.
        # This probably means someone ran Octopus in 'unocc' mode to
        # get eigenvalues (e.g. for band structures), and the values in
        # static/info will be the old (selfconsistent) ones.
        eigpath = directory / 'static/eigenvalues'
        if eigpath.is_file():
            with open(eigpath) as fd:
                kpts, eigs, occs = read_eigenvalues_file(fd)
                kpt_weights = np.ones(len(kpts))  # XXX ?  Or 1 / len(kpts) ?
                # XXX New Octopus probably has symmetry reduction !!
            results.update(eigenvalues=eigs, occupations=occs,
                           ibz_kpoints=kpts,
                           kpoint_weights=kpt_weights)
        return results

    def execute(self, directory, profile):
        profile.run(directory, None, self.outputname,
                    errorfile=self.errorname)

    def write_input(self, profile, directory, atoms, parameters, properties):
        txt = generate_input(atoms, process_special_kwargs(atoms, parameters))
        inp = directory / 'inp'
        inp.write_text(txt)

    def load_profile(self, cfg, **kwargs):
        return OctopusProfile.from_config(cfg, self.name, **kwargs)


class Octopus(GenericFileIOCalculator):
    """Octopus calculator.

    The label is always assumed to be a directory."""

    def __init__(self, profile=None, directory='.', **kwargs):
        """Create Octopus calculator.

        Label is always taken as a subdirectory.
        Restart is taken to be a label."""

        super().__init__(profile=profile,
                         template=OctopusTemplate(),
                         directory=directory,
                         parameters=kwargs)

    @classmethod
    def recipe(cls, **kwargs):
        from ase import Atoms
        system = Atoms()
        calc = Octopus(CalculationMode='recipe', **kwargs)
        system.calc = calc
        try:
            system.get_potential_energy()
        except OctopusIOError:
            pass
        else:
            raise OctopusIOError('Expected recipe, but found '
                                 'useful physical output!')



import numpy as np
import pytest

from ase.build import bulk


def get_spos(atoms):
    return atoms.get_scaled_positions(wrap=False)


@pytest.fixture()
def atoms():
    rng = np.random.RandomState(0)
    atoms = bulk('Ti') * (2, 2, 1)
    atoms.cell *= 0.9 + 0.2 * rng.random((3, 3))
    atoms.rattle(stdev=0.05, rng=rng)
    return atoms


@pytest.fixture()
def displacement(atoms):
    rng = np.random.RandomState(12345)
    return 0.1 * (rng.random((len(atoms), 3)) - 0.5)


@pytest.fixture()
def reference(displacement, atoms):
    return displacement + get_spos(atoms)


def test_abc_and_scaled_position(atoms):
    scaled = get_spos(atoms)
    for i, atom in enumerate(atoms):
        assert np.allclose(scaled[i], atom.scaled_position)
        assert np.allclose(scaled[i], [atom.a, atom.b, atom.c])


def test_set_scaled_position(atoms, displacement, reference):
    for i, atom in enumerate(atoms):
        atom.scaled_position += displacement[i]

    assert np.allclose(get_spos(atoms), reference)


def test_set_abc(atoms, displacement, reference):
    for i, atom in enumerate(atoms):
        atom.a += displacement[i, 0]
        atom.b += displacement[i, 1]
        atom.c += displacement[i, 2]

    assert np.allclose(get_spos(atoms), reference)



"""Window for setting up Carbon nanotubes and similar tubes.
"""


import ase.gui.ui as ui
from ase.build import nanotube
from ase.gui.i18n import _
from ase.gui.widgets import Element, pybutton

introtext = _("""\
Set up a Carbon nanotube by specifying the (n,m) roll-up vector.
Please note that m <= n.

Nanotubes of other elements can be made by specifying the element
and bond length.\
""")

py_template = """\
from ase.build import nanotube
atoms = nanotube({n}, {m}, length={length}, bond={bl:.3f}, symbol='{symb}')
"""

label_template = _('{natoms} atoms, diameter: {diameter:.3f} Å, '
                   'total length: {total_length:.3f} Å')


class SetupNanotube:
    """Window for setting up a (Carbon) nanotube."""

    def __init__(self, gui):
        self.element = Element('C', self.make)
        self.bondlength = ui.SpinBox(1.42, 0.0, 10.0, 0.01, self.make)
        self.n = ui.SpinBox(5, 1, 100, 1, self.make)
        self.m = ui.SpinBox(5, 0, 100, 1, self.make)
        self.length = ui.SpinBox(1, 1, 100, 1, self.make)
        self.description = ui.Label('')

        win = self.win = ui.Window(_('Nanotube'), wmtype='utility')
        win.add(ui.Text(introtext))
        win.add(self.element)
        win.add([_('Bond length: '),
                 self.bondlength,
                 _('Å')])
        win.add(_('Select roll-up vector (n,m) and tube length:'))
        win.add(['n:', self.n,
                 'm:', self.m,
                 _('Length:'), self.length])
        win.add(self.description)
        win.add([pybutton(_('Creating a nanoparticle.'), self.make),
                 ui.Button(_('Apply'), self.apply),
                 ui.Button(_('OK'), self.ok)])

        self.gui = gui
        self.atoms = None

    def make(self, element=None):
        symbol = self.element.symbol
        if symbol is None:
            self.atoms = None
            self.python = None
            self.description.text = ''
            return None

        n = self.n.value
        m = self.m.value
        length = self.length.value
        bl = self.bondlength.value
        self.atoms = nanotube(n, m, length=length, bond=bl, symbol=symbol)
        label = label_template.format(
            natoms=len(self.atoms),
            total_length=self.atoms.cell[2, 2],
            diameter=self.atoms.cell[0, 0] / 2)
        self.description.text = label
        return py_template.format(n=n, m=m, length=length, symb=symbol, bl=bl)

    def apply(self):
        self.make()
        if self.atoms is not None:
            self.gui.new_atoms(self.atoms)
            return True
        else:
            ui.error(_('No valid atoms.'),
                     _('You have not (yet) specified a consistent '
                       'set of parameters.'))
            return False

    def ok(self, *args):
        if self.apply():
            self.win.close()



import pytest

from ase.io.cif import CIFBlock, CIFLoop, parse_loop


def test_parse_cifloop_simple():
    dct = parse_loop(['_apples',
                      '_verysmallrocks',
                      '2 200',
                      '3 300',
                      '4 400'][::-1])
    assert dct['_apples'] == [2, 3, 4]
    assert dct['_verysmallrocks'] == [200, 300, 400]


def test_parse_cifloop_warn_duplicate_header():
    with pytest.warns(UserWarning):
        parse_loop(['_hello', '_hello'])


def test_parse_cifloop_incomplete():
    with pytest.raises(RuntimeError):
        parse_loop(['_spam', '_eggs', '1 2', '1'][::-1])


def test_cifloop_roundtrip():
    loop = CIFLoop()
    loop.add('_potatoes', [2.5, 3.0, -1.0], '{:8.5f}')
    loop.add('_eggs', [1, 2, 3], '{:2d}')
    string = loop.tostring() + '\n'
    print('hmm', string)
    lines = string.splitlines()[::-1]
    assert lines.pop() == 'loop_'

    for line in lines:
        print(repr(line))
    dct = parse_loop(lines)
    assert dct['_potatoes'] == pytest.approx([2.5, 3.0, -1.0])
    assert dct['_eggs'] == [1, 2, 3]


@pytest.fixture()
def cifblock():
    return CIFBlock('hello', {'_cifkey': 42})


def test_repr(cifblock):
    text = repr(cifblock)
    assert 'hello' in text
    assert '_cifkey' in text


def test_mapping(cifblock):
    assert len(cifblock) == 1
    assert len(list(cifblock)) == 1


def test_various(cifblock):
    assert cifblock.get_cellpar() is None
    assert cifblock.get_cell().rank == 0


def test_deuterium():
    # Verify that the symbol 'D' becomes hydrogen ('H') with mass 2(-ish).
    symbols = ['H', 'D', 'D', 'He']
    block = CIFBlock('deuterium', dict(_atom_site_type_symbol=symbols))
    assert block.get_symbols() == ['H', 'H', 'H', 'He']
    masses = block._get_masses()
    assert all(masses.round().astype(int) == [1, 2, 2, 4])



import os

import pytest

from ase import Atoms
from ase.build import molecule
from ase.calculators.emt import EMT
from ase.db import connect


@pytest.fixture(scope='module')
def url(mysql_port):
    pytest.importorskip('pymysql')

    on_ci_server = 'CI_PROJECT_DIR' in os.environ

    if on_ci_server:
        # CI server configured to use non-standard port 3307
        # instead of the default 3306 port. This is to test
        # for proper passing of the port for creating the
        # mysql connection
        db_url = f'mysql://root:ase@mysql:{mysql_port}/testase_mysql'
        # HOST = 'mysql'
        # USER = 'root'
        # PASSWD = 'ase'
        # DB_NAME = 'testase_mysql'
    else:
        db_url = os.environ.get('MYSQL_DB_URL')
        # HOST = os.environ.get('MYSQL_HOST', None)
        # USER = os.environ.get('MYSQL_USER', None)
        # PASSWD = os.environ.get('MYSQL_PASSWD', None)
        # DB_NAME = os.environ.get('MYSQL_DB_NAME', None)

    if db_url is None:
        msg = ('Not on GitLab CI server. To run this test '
               'host, username, password and database name '
               'must be in the environment variables '
               'MYSQL_HOST, MYSQL_USER, MYSQL_PASSWD and '
               'MYSQL_DB_NAME, respectively.')
        pytest.skip(msg)
    return db_url


@pytest.fixture()
def db(url):
    return connect(url)


@pytest.fixture()
def h2o():
    return molecule('H2O')


def test_connect(db):
    db.delete([row.id for row in db.select()])


def test_write_read(db):
    co = Atoms('CO', positions=[(0, 0, 0), (0, 0, 1.1)])
    uid = db.write(co, tag=1, type='molecule')

    co_db = db.get(id=uid)
    atoms_db = co_db.toatoms()

    assert len(atoms_db) == 2
    assert atoms_db[0].symbol == co[0].symbol
    assert atoms_db[1].symbol == co[1].symbol
    assert co_db.tag == 1
    assert co_db.type == 'molecule'


def test_write_read_with_calculator(db, h2o):
    calc = EMT(dummy_param=2.4)
    h2o.calc = calc

    uid = db.write(h2o)

    h2o_db = db.get(id=uid).toatoms()

    # Back in the days we allowed reconstructing calculators.
    # For security we don't anymore.
    assert h2o_db.calc is None

    # Check that get_atoms function works
    db.get_atoms(H=2)
    # XXX We should assert something should we not?


def test_update(db, h2o):
    h2o = molecule('H2O')

    uid = db.write(h2o, type='molecule')
    db.update(id=uid, type='oxide')

    atoms_type = db.get(id=uid).type

    assert atoms_type == 'oxide'


def test_delete(db, h2o):
    h2o = molecule('H2O')
    uid = db.write(h2o, type='molecule')

    # Make sure that we can get the value
    db.get(id=uid)
    db.delete([uid])

    with pytest.raises(KeyError):
        db.get(id=uid)


def test_read_write_bool_key_value_pair(db, h2o):
    # Make sure we can read and write boolean key value pairs
    uid = db.write(h2o, is_water=True, is_solid=False)
    row = db.get(id=uid)
    assert row.is_water
    assert not row.is_solid



import numpy as np

from ase.build import bulk, molecule
from ase.calculators.gulp import GULP
from ase.filters import FrechetCellFilter
from ase.optimize import BFGS


def test_gulp_opt(gulp_factory):
    # GULP optmization test
    atoms = molecule('H2O')
    atoms1 = atoms.copy()
    atoms1.calc = GULP(library='reaxff.lib')
    with BFGS(atoms1) as opt1:
        opt1.run(fmax=0.005)

    atoms2 = atoms.copy()
    calc2 = GULP(keywords='opti conp', library='reaxff.lib')
    with calc2.get_optimizer(atoms2) as opt2:
        opt2.run()

    print(np.abs(opt1.atoms.positions - opt2.atoms.positions))
    assert np.abs(opt1.atoms.positions - opt2.atoms.positions).max() < 1e-5

    # GULP optimization test using stress
    atoms = bulk('Au', 'bcc', a=2.7, cubic=True)
    atoms1 = atoms.copy()
    atoms1.calc = GULP(keywords='conp gradient stress_out',
                       library='reaxff_general.lib')
    atoms1f = FrechetCellFilter(atoms1)
    with BFGS(atoms1f) as opt1:
        opt1.run(fmax=0.005)

    atoms2 = atoms.copy()
    calc2 = GULP(keywords='opti conp', library='reaxff_general.lib')
    with calc2.get_optimizer(atoms2) as opt2:
        opt2.run()

    print(np.abs(opt1.atoms.positions - opt2.atoms.positions))
    assert np.abs(opt1.atoms.positions - opt2.atoms.positions).max() < 1e-5



from random import randint

import numpy as np
import pytest

from ase import Atoms
from ase.build import bulk
from ase.spacegroup import crystal, spacegroup
from ase.utils.structure_comparator import (
    SpgLibNotFoundError,
    SymmetryEquivalenceCheck,
)

heavy_test = False


@pytest.fixture(scope='module')
def comparator():
    return SymmetryEquivalenceCheck()


def get_atoms_with_mixed_elements(crystalstructure="fcc"):
    atoms = bulk("Al", crystalstructure=crystalstructure, a=3.2)
    atoms = atoms * (2, 2, 2)
    symbs = ["Al", "Cu", "Zn"]
    symbols = [symbs[randint(0, len(symbs) - 1)] for _ in range(len(atoms))]
    for i in range(len(atoms)):
        atoms[i].symbol = symbols[i]
    return atoms


def test_compare(comparator):
    s1 = bulk("Al")
    s1 = s1 * (2, 2, 2)
    s2 = bulk("Al")
    s2 = s2 * (2, 2, 2)
    assert comparator.compare(s1, s2)


def test_fcc_bcc(comparator):
    s1 = bulk("Al", crystalstructure="fcc")
    s2 = bulk("Al", crystalstructure="bcc", a=4.05)
    s1 = s1 * (2, 2, 2)
    s2 = s2 * (2, 2, 2)
    assert not comparator.compare(s1, s2)


def test_single_impurity(comparator):
    s1 = bulk("Al")
    s1 = s1 * (2, 2, 2)
    s1[0].symbol = "Mg"
    s2 = bulk("Al")
    s2 = s2 * (2, 2, 2)
    s2[3].symbol = "Mg"
    assert comparator.compare(s1, s2)


def test_translations(comparator):
    s1 = get_atoms_with_mixed_elements()
    s2 = s1.copy()

    xmax = 2.0 * np.max(s1.get_cell().T)
    N = 3
    dx = xmax / N
    pos_ref = s2.get_positions()
    structures = []
    for i in range(N):
        for j in range(N):
            for k in range(N):
                displacement = np.array([dx * i, dx * j, dx * k])
                new_pos = pos_ref + displacement
                s2.set_positions(new_pos)
                structures.append(s2)
    assert comparator.compare(s1, structures)


def test_rot_60_deg(comparator):
    s1 = get_atoms_with_mixed_elements()
    s2 = s1.copy()
    ca = np.cos(np.pi / 3.0)
    sa = np.sin(np.pi / 3.0)
    matrix = np.array([[ca, sa, 0.0], [-sa, ca, 0.0], [0.0, 0.0, 1.0]])
    s2.set_positions(matrix.dot(s2.get_positions().T).T)
    s2.set_cell(matrix.dot(s2.get_cell().T).T)
    assert comparator.compare(s1, s2)


def test_rot_120_deg(comparator):
    s1 = get_atoms_with_mixed_elements()
    s2 = s1.copy()
    ca = np.cos(2.0 * np.pi / 3.0)
    sa = np.sin(2.0 * np.pi / 3.0)
    matrix = np.array([[ca, sa, 0.0], [-sa, ca, 0.0], [0.0, 0.0, 1.0]])
    s2.set_positions(matrix.dot(s2.get_positions().T).T)
    s2.set_cell(matrix.dot(s2.get_cell().T).T)
    assert comparator.compare(s1, s2)


def test_rotations_to_standard(comparator):
    s1 = Atoms("Al")
    tol = 1E-6
    num_tests = 4
    if heavy_test:
        num_tests = 20
    for _ in range(num_tests):
        cell = np.random.rand(3, 3) * 4.0 - 4.0
        s1.set_cell(cell)
        new_cell = comparator._standarize_cell(s1).get_cell().T
        assert abs(new_cell[1, 0]) < tol
        assert abs(new_cell[2, 0]) < tol
        assert abs(new_cell[2, 1]) < tol


def test_point_inversion(comparator):
    s1 = get_atoms_with_mixed_elements()
    s2 = s1.copy()
    s2.set_positions(-s2.get_positions())
    assert comparator.compare(s1, s2)


def test_mirror_plane(comparator):
    s1 = get_atoms_with_mixed_elements(crystalstructure="hcp")
    s2 = s1.copy()
    mat = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, -1.0]])
    s2.set_positions(mat.dot(s2.get_positions().T).T)
    assert comparator.compare(s1, s2)

    mat = np.array([[-1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    s2.set_positions(mat.dot(s1.get_positions().T).T)
    assert comparator.compare(s1, s2)

    mat = np.array([[1.0, 0.0, 0.0], [0.0, -1.0, 0.0], [0.0, 0.0, 1.0]])
    s2.set_positions(mat.dot(s1.get_positions().T).T)
    assert comparator.compare(s1, s2)


def test_hcp_symmetry_ops(comparator):
    s1 = get_atoms_with_mixed_elements(crystalstructure="hcp")
    s2 = s1.copy()
    sg = spacegroup.Spacegroup(194)
    cell = s2.get_cell().T
    inv_cell = np.linalg.inv(cell)
    operations = sg.get_rotations()
    if not heavy_test:
        operations = operations[::int(np.ceil(len(operations) / 4))]
    for op in operations:
        s1 = get_atoms_with_mixed_elements(crystalstructure="hcp")
        s2 = s1.copy()
        transformed_op = cell.dot(op).dot(inv_cell)
        s2.set_positions(transformed_op.dot(s1.get_positions().T).T)
        assert comparator.compare(s1, s2)


def test_fcc_symmetry_ops(comparator):
    s1 = get_atoms_with_mixed_elements()
    s2 = s1.copy()
    sg = spacegroup.Spacegroup(225)
    cell = s2.get_cell().T
    inv_cell = np.linalg.inv(cell)
    operations = sg.get_rotations()
    if not heavy_test:
        operations = operations[::int(np.ceil(len(operations) / 4))]
    for op in operations:
        s1 = get_atoms_with_mixed_elements()
        s2 = s1.copy()
        transformed_op = cell.dot(op).dot(inv_cell)
        s2.set_positions(transformed_op.dot(s1.get_positions().T).T)
        assert comparator.compare(s1, s2)


def test_bcc_symmetry_ops(comparator):
    s1 = get_atoms_with_mixed_elements(crystalstructure="bcc")
    s2 = s1.copy()
    sg = spacegroup.Spacegroup(229)
    cell = s2.get_cell().T
    inv_cell = np.linalg.inv(cell)
    operations = sg.get_rotations()
    if not heavy_test:
        operations = operations[::int(np.ceil(len(operations) / 4))]
    for op in operations:
        s1 = get_atoms_with_mixed_elements(crystalstructure="bcc")
        s2 = s1.copy()
        transformed_op = cell.dot(op).dot(inv_cell)
        s2.set_positions(transformed_op.dot(s1.get_positions().T).T)
        assert comparator.compare(s1, s2)


def test_bcc_translation(comparator):
    s1 = get_atoms_with_mixed_elements(crystalstructure="bcc")
    s2 = s1.copy()
    s2.set_positions(s2.get_positions() + np.array([6.0, -2.0, 1.0]))
    assert comparator.compare(s1, s2)


def test_one_atom_out_of_pos(comparator):
    s1 = get_atoms_with_mixed_elements()
    s2 = s1.copy()
    pos = s1.get_positions()
    pos[0, :] += 0.2
    s2.set_positions(pos)
    assert not comparator.compare(s1, s2)


def test_reduce_to_primitive(comparator):
    atoms1 = crystal(symbols=['V', 'Li', 'O'],
                     basis=[(0.000000, 0.000000, 0.000000),
                            (0.333333, 0.666667, 0.000000),
                            (0.333333, 0.000000, 0.250000)],
                     spacegroup=167,
                     cellpar=[5.123, 5.123, 13.005, 90., 90., 120.],
                     size=[1, 1, 1], primitive_cell=False)

    atoms2 = crystal(symbols=['V', 'Li', 'O'],
                     basis=[(0.000000, 0.000000, 0.000000),
                            (0.333333, 0.666667, 0.000000),
                            (0.333333, 0.000000, 0.250000)],
                     spacegroup=167,
                     cellpar=[5.123, 5.123, 13.005, 90., 90., 120.],
                     size=[1, 1, 1], primitive_cell=True)
    try:
        # Tell the comparator to reduce to primitive cell
        comparator.to_primitive = True

        assert comparator.compare(atoms1, atoms2)
    except SpgLibNotFoundError:
        pass

    # Reset the comparator to its original state
    comparator.to_primitive = False


def test_order_of_candidates(comparator):
    s1 = bulk("Al", crystalstructure='fcc', a=3.2)
    s1 = s1 * (2, 2, 2)
    s2 = s1.copy()
    s1.positions[0, :] += .2

    assert comparator.compare(s2, s1) == comparator.compare(s1, s2)


def test_original_paper_structures():
    # Structures from the original paper:
    # Comput. Phys. Commun. 183, 690-697 (2012)
    # They should evaluate equal (within a certain tolerance)
    syms = ['O', 'O', 'Mg', 'F']
    cell1 = [(3.16, 0.00, 0.00), (-0.95, 4.14, 0.00), (-0.95, -0.22, 4.13)]
    p1 = [(0.44, 0.40, 0.30), (0.94, 0.40, 0.79),
          (0.45, 0.90, 0.79), (0.94, 0.40, 0.29)]
    s1 = Atoms(syms, cell=cell1, scaled_positions=p1, pbc=True)

    cell2 = [(6.00, 0.00, 0.00), (1.00, 3.00, 0.00), (2.00, -3.00, 3.00)]
    p2 = [(0.00, 0.00, 0.00), (0.00, 0.00, 0.50),
          (0.50, 0.00, 0.00), (0.00, 0.50, 0.00)]
    s2 = Atoms(syms, cell=cell2, scaled_positions=p2, pbc=True)

    comp = SymmetryEquivalenceCheck()

    assert comp.compare(s1, s2)
    assert comp.compare(s2, s1) == comp.compare(s1, s2)


def test_symmetrical_one_element_out(comparator):
    s1 = get_atoms_with_mixed_elements()
    s1.set_chemical_symbols(['Zn', 'Zn', 'Al', 'Zn', 'Zn', 'Al', 'Zn', 'Zn'])
    s2 = s1.copy()
    s2.positions[0, :] += 0.2
    assert not comparator.compare(s1, s2)
    assert not comparator.compare(s2, s1)


def test_one_vs_many():
    s1 = Atoms('H3', positions=[[0.5, 0.5, 0], [0.5, 1.5, 0], [1.5, 1.5, 0]],
               cell=[2, 2, 2], pbc=True)
    # Get the unit used for position comparison
    u = (s1.get_volume() / len(s1))**(1 / 3)
    comp = SymmetryEquivalenceCheck(stol=.095 / u, scale_volume=True)
    s2 = s1.copy()
    assert comp.compare(s1, s2)
    s2_list = []
    s3 = Atoms('H3', positions=[[0.5, 0.5, 0], [0.5, 1.5, 0], [1.5, 1.5, 0]],
               cell=[3, 3, 3], pbc=True)
    s2_list.append(s3)
    for d in np.linspace(0.1, 1.0, 5):
        s2 = s1.copy()
        s2.positions[0] += [d, 0, 0]
        s2_list.append(s2)
    assert not comp.compare(s1, s2_list[:-1])
    assert comp.compare(s1, s2_list)


def test_supercell_w_periodic_atom_removed(comparator):
    s1 = Atoms(['H', 'H', 'He', 'He', 'He'], positions=[[.1, .1, .1],
                                                        [-.1, -.1, -.1],
                                                        [.4, .3, .2],
                                                        [.3, .6, .3],
                                                        [.8, .5, .6]],
               cell=[1, 1, 1], pbc=True)
    s1 *= (2, 1, 1)
    a0 = s1.copy()
    del a0[0]

    a5 = s1.copy()
    del a5[5]

    assert comparator.compare(a0, a5)
    assert comparator.compare(a5, a0) == comparator.compare(a0, a5)



# type: ignore
import os.path

import numpy as np
import pytest

from ase import Atoms
from ase.calculators.turbomole import Turbomole


@pytest.fixture(scope="function")
def atoms():
    return Atoms('H2', positions=[(0, 0, 0), (0, 0, 1.1)])


def test_turbomole_H2_rhf_singlet(atoms, turbomole_factory):
    # Write all commands for the define command in a string
    define_str = '\n\na coord\n*\nno\nb all sto-3g hondo\n*\neht\n\n\n\n*'

    atoms.calc = Turbomole(define_str=define_str)

    # Run turbomole
    assert np.isclose(atoms.get_potential_energy(), -28.205659, atol=1e-5)


def test_turbomole_H2_uhf_singlet(atoms, turbomole_factory):
    atoms.calc = Turbomole(**{
        "multiplicity": 1, "uhf": True, "use dft": True
    })

    # Run turbomole
    assert np.isclose(atoms.get_potential_energy(), -30.828865, atol=1e-5)

    # check that it performed a DFT calculation (i.e. basic inputs were most
    # likely understood, cf. issue #735)
    dft_in_output = False
    with open("ASE.TM.dscf.out") as fd:
        for line in fd:
            if "density functional" in line:
                dft_in_output = True
    assert dft_in_output

    # also check that UHF was understood (alpha and beta files present)
    assert os.path.exists("alpha")
    assert os.path.exists("beta")
    assert not os.path.exists("mos")



from ase import Atoms
from ase.calculators.test import FreeElectrons
from ase.cell import Cell


def test_monoclinic():
    """Test band structure from different variations of hexagonal cells."""
    mc1 = Cell([[1, 0, 0], [0, 1, 0], [0, 0.2, 1]])
    par = mc1.cellpar()
    mc2 = Cell.new(par)
    mc3 = Cell([[1, 0, 0], [0, 1, 0], [-0.2, 0, 1]])
    mc4 = Cell([[1, 0, 0], [-0.2, 1, 0], [0, 0, 1]])
    path = 'GYHCEM1AXH1'

    firsttime = True
    for cell in [mc1, mc2, mc3, mc4]:
        a = Atoms(cell=cell, pbc=True)
        a.cell *= 3
        a.calc = FreeElectrons(nvalence=1, kpts={'path': path})

        lat = a.cell.get_bravais_lattice()
        assert lat.name == 'MCL'
        a.get_potential_energy()
        bs = a.calc.band_structure()
        coords, labelcoords, labels = bs.get_labels()
        assert ''.join(labels) == path
        e_skn = bs.energies

        if firsttime:
            coords1 = coords
            labelcoords1 = labelcoords
            e_skn1 = e_skn
            firsttime = False
        else:
            for d in [coords - coords1,
                      labelcoords - labelcoords1,
                      e_skn - e_skn1]:
                print(abs(d).max())
                assert abs(d).max() < 1e-13, d



from ase.build import bulk
from ase.calculators.dmol import DMol3


def test_Al_dmol(dmol_factory):
    atoms = bulk('Al')
    calc = DMol3()
    atoms.calc = calc
    atoms.get_potential_energy()
    atoms.get_forces()



# Note:
# Try to avoid module level import statements here to reduce
# import time during CLI execution
import sys

from ase.cli.main import CLIError

template_help = """
Without argument, looks for ~/.ase/template.py.  Otherwise,
expects the comma separated list of the fields to include
in their left-to-right order.  Optionally, specify the
lexicographical sort hierarchy (0 is outermost sort) and if the
sort should be ascending or descending (1 or -1).  By default,
sorting is descending, which makes sense for most things except
index (and rank, but one can just sort by the thing which is
ranked to get ascending ranks).

* example: ase diff start.cif stop.cif --template
* i:0:1,el,dx,dy,dz,d,rd

possible fields:

*    i: index
*    dx,dy,dz,d: displacement/displacement components
*    dfx,dfy,dfz,df: difference force/force components
*    afx,afy,afz,af: average force/force components
*    p1x,p1y,p1z,p: first image positions/position components
*    p2x,p2y,p2z,p: second image positions/position components
*    f1x,f1y,f1z,f: first image forces/force components
*    f2x,f2y,f2z,f: second image forces/force components
*    an: atomic number
*    el: atomic element
*    t: atom tag
*    r<col>: the rank of that atom with respect to the column

It is possible to change formatters in the template file."""


class CLICommand:
    """Print differences between atoms/calculations.

    Supports taking differences between different calculation runs of
    the same system as well as neighboring geometric images for one
    calculation run of a system. As part of a difference table or as a
    standalone display table, fields for non-difference quantities of image 1
    and image 2 are also provided.

    See the --template-help for the formatting exposed in the CLI.  More
    customization requires changing the input arguments to the Table
    initialization and/or editing the templates file.
    """

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument
        add('file',
            help="""Possible file entries are

    * 2 non-trajectory files: difference between them
    * 1 trajectory file: difference between consecutive images
    * 2 trajectory files: difference between corresponding image numbers
    * 1 trajectory file followed by hyphen-minus (ASCII 45): for display

    Note deltas are defined as 2 - 1.

    Use [FILE]@[SLICE] to select images.
                    """,
            nargs='+')
        add('-r',
            '--rank-order',
            metavar='FIELD',
            nargs='?',
            const='d',
            type=str,
            help="""Order atoms by rank, see --template-help for possible
fields.

The default value, when specified, is d.  When not
specified, ordering is the same as that provided by the
generator.  For hierarchical sorting, see template.""")
        add('-c', '--calculator-outputs', action="store_true",
            help="display calculator outputs of forces and energy")
        add('--max-lines', metavar='N', type=int,
            help="show only so many lines (atoms) in each table "
            ", useful if rank ordering")
        add('-t', '--template', metavar='TEMPLATE', nargs='?', const='rc',
            help="""See --template-help for the help on this option.""")
        add('--template-help', help="""Prints the help for the template file.
                Usage `ase diff - --template-help`""", action="store_true")
        add('-s', '--summary-functions', metavar='SUMFUNCS', nargs='?',
            help="""Specify the summary functions.
            Possible values are `rmsd` and `dE`.
            Comma separate more than one summary function.""")
        add('--log-file', metavar='LOGFILE', help="print table to file")
        add('--as-csv', action="store_true",
            help="output table in csv format")
        add('--precision', metavar='PREC',
            default=2, type=int,
            help="precision used in both display and sorting")

    @staticmethod
    def run(args, parser):
        import io

        if args.template_help:
            print(template_help)
            return

        encoding = 'utf-8'

        if args.log_file is None:
            out = io.TextIOWrapper(sys.stdout.buffer, encoding=encoding)
        else:
            out = open(args.log_file, 'w', encoding=encoding)

        with out:
            CLICommand.diff(args, out)

    @staticmethod
    def diff(args, out):
        from ase.cli.template import (
            Table,
            TableFormat,
            energy_delta,
            field_specs_on_conditions,
            rmsd,
            slice_split,
            summary_functions_on_conditions,
        )
        from ase.io import read

        if args.template is None:
            field_specs = field_specs_on_conditions(
                args.calculator_outputs, args.rank_order)
        else:
            field_specs = args.template.split(',')
            if not args.calculator_outputs:
                for field_spec in field_specs:
                    if 'f' in field_spec:
                        raise CLIError(
                            "field requiring calculation outputs "
                            "without --calculator-outputs")

        if args.summary_functions is None:
            summary_functions = summary_functions_on_conditions(
                args.calculator_outputs)
        else:
            summary_functions_dct = {
                'rmsd': rmsd,
                'dE': energy_delta}
            summary_functions = args.summary_functions.split(',')
            if not args.calculator_outputs:
                for sf in summary_functions:
                    if sf == 'dE':
                        raise CLIError(
                            "summary function requiring calculation outputs "
                            "without --calculator-outputs")
            summary_functions = [summary_functions_dct[i]
                                 for i in summary_functions]

        have_two_files = len(args.file) == 2
        file1 = args.file[0]
        actual_filename, index = slice_split(file1)
        atoms1 = read(actual_filename, index)
        natoms1 = len(atoms1)

        if have_two_files:
            if args.file[1] == '-':
                atoms2 = atoms1

                def header_fmt(c):
                    return f'image # {c}'
            else:
                file2 = args.file[1]
                actual_filename, index = slice_split(file2)
                atoms2 = read(actual_filename, index)
                natoms2 = len(atoms2)

                same_length = natoms1 == natoms2
                one_l_one = natoms1 == 1 or natoms2 == 1

                if not same_length and not one_l_one:
                    raise CLIError(
                        "Trajectory files are not the same length "
                        "and both > 1\n{}!={}".format(
                            natoms1, natoms2))
                elif not same_length and one_l_one:
                    print(
                        "One file contains one image "
                        "and the other multiple images,\n"
                        "assuming you want to compare all images "
                        "with one reference image")
                    if natoms1 > natoms2:
                        atoms2 = natoms1 * atoms2
                    else:
                        atoms1 = natoms2 * atoms1

                    def header_fmt(c):
                        return f'sys-ref image # {c}'
                else:
                    def header_fmt(c):
                        return f'sys2-sys1 image # {c}'
        else:
            atoms2 = atoms1.copy()
            atoms1 = atoms1[:-1]
            atoms2 = atoms2[1:]
            natoms2 = natoms1 = natoms1 - 1

            def header_fmt(c):
                return f'images {c + 1}-{c}'

        natoms = natoms1  # = natoms2

        output = ''
        tableformat = TableFormat(precision=args.precision,
                                  columnwidth=7 + args.precision)

        table = Table(
            field_specs,
            max_lines=args.max_lines,
            tableformat=tableformat,
            summary_functions=summary_functions)

        for counter in range(natoms):
            table.title = header_fmt(counter)
            output += table.make(atoms1[counter],
                                 atoms2[counter], csv=args.as_csv) + '\n'
        print(output, file=out)



import numpy as np
import pytest

from ase.build import molecule
from ase.calculators.emt import EMT
from ase.constraints import FixedLine, FixedPlane
from ase.optimize import BFGS


@pytest.fixture(params=[FixedLine, FixedPlane])
def fixture_test_class(request):
    return request.param


@pytest.mark.parametrize(
    'indices', [
        0,
        [0],
        [0, 1],
        np.array([0, 1], dtype=np.int64),
    ]
)
def test_valid_inputs_indices(fixture_test_class, indices):
    _ = fixture_test_class(indices, [1, 0, 0])


@pytest.mark.parametrize(
    'indices', [
        [0, 1, 1],
        [[0, 1], [0, 1]],
    ]
)
def test_invalid_inputs_indices(fixture_test_class, indices):
    with pytest.raises(ValueError) as _:
        _ = fixture_test_class(indices, [1, 0, 0])


@pytest.mark.parametrize('direction', [[0, 0, 1], (0, 0, 1)])
def test_valid_inputs_direction(fixture_test_class, direction):
    _ = fixture_test_class(0, direction)


@pytest.mark.parametrize('direction', [[0, 1], None, "42"])
def test_invalid_inputs_direction(fixture_test_class, direction):
    with pytest.raises(Exception) as _:
        _ = FixedLine(0, direction)


def _check_simple_constraints(constraints, indices):
    mol = molecule("butadiene")
    mol.set_constraint(constraints)

    assert len(mol.constraints) == 1
    assert isinstance(constraints.dir, np.ndarray)
    assert (np.asarray([1, 0, 0]) == constraints.dir).all()

    mol.calc = EMT()

    cold_positions = mol[indices].positions.copy()
    opt = BFGS(mol)
    opt.run(steps=5)
    cnew_positions = mol[indices].positions.copy()

    return cold_positions, cnew_positions


@pytest.mark.parametrize('indices', [0, [0], [0, 1]])
def test_repr_fixedline(fixture_test_class, indices):
    repr(FixedLine(indices, [1, 0, 0])) == (
        "<FixedLine: "
        "{'indices': " + str(indices) + ", 'direction': [1. 0. 0.]}>"
    )


@pytest.mark.parametrize(
    'indices,expected', [
        (0, 2),
        ([0], 2),
        ([0, 1], 4),
    ]
)
def test_removed_dof_fixedline(indices, expected):
    mol = molecule("butadiene")  # `get_removed_dof` requires an `Atoms` object
    constraints = FixedLine(indices, direction=[1, 0, 0])
    assert constraints.get_removed_dof(atoms=mol) == expected


@pytest.mark.optimize()
@pytest.mark.parametrize('indices', [[0], [0, 1]])
def test_constrained_optimization_fixedline(indices):
    """
    A single int is not tested as that changes the call from Atoms.positions
    to Atom.position
    """
    constraints = FixedLine(indices, [1, 0, 0])

    cold_positions, cnew_positions = _check_simple_constraints(
        constraints, indices
    )
    assert np.max(np.abs(cnew_positions[:, 1:] - cold_positions[:, 1:])) < 1e-8
    assert np.max(np.abs(cnew_positions[:, 0] - cold_positions[:, 0])) > 1e-8


@pytest.mark.parametrize('indices', [0, [0], [0, 1]])
def test_repr_fixedplane(fixture_test_class, indices):
    repr(FixedPlane(indices, [1, 0, 0])) == (
        "<FixedPlane: "
        "{'indices': " + str(indices) + ", 'direction': [1. 0. 0.]}>"
    )


@pytest.mark.parametrize(
    'indices,expected', [
        (0, 1),
        ([0], 1),
        ([0, 1], 2),
    ]
)
def test_removed_dof_fixedplane(indices, expected):
    mol = molecule("butadiene")  # `get_removed_dof` requires an `Atoms` object
    constraints = FixedPlane(indices, direction=[1, 0, 0])
    assert constraints.get_removed_dof(atoms=mol) == expected


@pytest.mark.optimize()
@pytest.mark.parametrize('indices', [[0], [0, 1]])
def test_constrained_optimization_fixedplane(indices):
    """
    A single int is not tested as that changes the call from Atoms.positions
    to Atom.position
    """
    constraints = FixedPlane(indices, [1, 0, 0])

    cold_positions, cnew_positions = _check_simple_constraints(
        constraints, indices
    )
    assert np.max(np.abs(cnew_positions[:, 1:] - cold_positions[:, 1:])) > 1e-8
    assert np.max(np.abs(cnew_positions[:, 0] - cold_positions[:, 0])) < 1e-8



"""
Check the many ways of specifying KPOINTS
"""
import os

import pytest

from ase.build import bulk
from ase.calculators.vasp.create_input import format_kpoints

from .filecmp_ignore_whitespace import filecmp_ignore_whitespace

calc = pytest.mark.calculator


@pytest.fixture()
def atoms():
    return bulk('Al', 'fcc', a=4.5, cubic=True)


def check_kpoints_line(n, contents):
    """Assert the contents of a line"""
    with open('KPOINTS') as fd:
        lines = fd.readlines()
    assert lines[n].strip() == contents


@pytest.fixture()
def write_kpoints(atoms):
    """Helper fixture to write the input kpoints file"""
    def _write_kpoints(factory, **kwargs):
        calc = factory.calc(**kwargs)
        calc.initialize(atoms)
        calc.write_kpoints(atoms=atoms)
        return atoms, calc

    return _write_kpoints


def test_vasp_kpoints_111(atoms):
    # Default to (1 1 1)
    string = format_kpoints(gamma=True, atoms=atoms, kpts=(1, 1, 1))
    check_kpoints_string(string, 2, 'Gamma')
    check_kpoints_string(string, 3, '1 1 1')


def test_vasp_kpoints_3_tuple(atoms):
    string = format_kpoints(gamma=False, kpts=(4, 4, 4), atoms=atoms)
    lines = string.split('\n')
    assert lines[1] == '0'
    assert lines[2] == 'Monkhorst-Pack'
    assert lines[3] == '4 4 4'


def check_kpoints_string(string, lineno, value):
    assert string.splitlines()[lineno].strip() == value


def test_vasp_kpoints_auto(atoms):
    string = format_kpoints(atoms=atoms, kpts=20)
    check_kpoints_string(string, 1, '0')
    check_kpoints_string(string, 2, 'Auto')
    check_kpoints_string(string, 3, '20')


def test_vasp_kpoints_1_element_list_gamma(atoms):
    # 1-element list ok, Gamma ok
    string = format_kpoints(atoms=atoms, kpts=[20], gamma=True)
    check_kpoints_string(string, 1, '0')
    check_kpoints_string(string, 2, 'Auto')
    check_kpoints_string(string, 3, '20')


@calc('vasp')
def test_kspacing_supress_kpoints_file(factory, write_kpoints):
    # KSPACING suppresses KPOINTS file
    Al, calc = write_kpoints(factory, kspacing=0.23)
    calc.write_incar(Al)
    assert not os.path.isfile('KPOINTS')
    with open('INCAR') as fd:
        assert 'KSPACING = 0.230000\n' in fd.readlines()


@calc('vasp')
def test_negative_kspacing_error(factory, write_kpoints):
    # Negative KSPACING raises an error
    with pytest.raises(ValueError):
        write_kpoints(factory, kspacing=-0.5)


def test_weighted(atoms, testdir):
    # Explicit weighted points with nested lists, Cartesian if not specified
    string = format_kpoints(
        atoms=atoms,
        kpts=[[0.1, 0.2, 0.3, 2], [0.0, 0.0, 0.0, 1],
              [0.0, 0.5, 0.5, 2]])

    with open('KPOINTS', 'w') as fd:
        fd.write(string)

    with open('KPOINTS.ref', 'w') as fd:
        fd.write("""KPOINTS created by Atomic Simulation Environment
    3 \n\
    Cartesian
    0.100000 0.200000 0.300000 2.000000 \n\
    0.000000 0.000000 0.000000 1.000000 \n\
    0.000000 0.500000 0.500000 2.000000 \n\
    """)

    assert filecmp_ignore_whitespace('KPOINTS', 'KPOINTS.ref')


def test_explicit_auto_weight(atoms, testdir):
    # Explicit points as list of tuples, automatic weighting = 1.
    string = format_kpoints(
        atoms=atoms,
        kpts=[(0.1, 0.2, 0.3), (0.0, 0.0, 0.0), (0.0, 0.5, 0.5)],
        reciprocal=True)

    with open('KPOINTS', 'w') as fd:
        fd.write(string)

    with open('KPOINTS.ref', 'w') as fd:
        fd.write("""KPOINTS created by Atomic Simulation Environment
    3 \n\
    Reciprocal
    0.100000 0.200000 0.300000 1.0 \n\
    0.000000 0.000000 0.000000 1.0 \n\
    0.000000 0.500000 0.500000 1.0 \n\
    """)

    assert filecmp_ignore_whitespace('KPOINTS', 'KPOINTS.ref')


def test_bandpath(atoms):
    bandpath = atoms.cell.bandpath('GXMGRX,MR', npoints=100)
    string = format_kpoints(atoms=atoms, kpts=bandpath.kpts, reciprocal=True)
    check_kpoints_string(string, 1, '100')
    check_kpoints_string(string, 2, 'Reciprocal')
    check_kpoints_string(string, 3, '0.000000 0.000000 0.000000 1.0')
    check_kpoints_string(string, 102, '0.500000 0.500000 0.500000 1.0')



"""
Implementation of the Precon abstract base class and subclasses
"""

import copy
import sys
import time
import warnings
from abc import ABC, abstractmethod

import numpy as np
from scipy import sparse
from scipy.interpolate import CubicSpline
from scipy.sparse.linalg import spsolve

import ase.units as units
import ase.utils.ff as ff
from ase.constraints import FixAtoms
from ase.filters import Filter
from ase.geometry import find_mic
from ase.neighborlist import neighbor_list
from ase.optimize.precon.neighbors import (
    estimate_nearest_neighbour_distance,
    get_neighbours,
)
from ase.utils import longsum, tokenize_version

try:
    import pyamg
except ImportError:
    have_pyamg = False
else:
    have_pyamg = True


def create_pyamg_solver(P, max_levels=15):
    if not have_pyamg:
        raise RuntimeError(
            'pyamg not available: install with `pip install pyamg`')
    filter_key = 'filter'
    if tokenize_version(pyamg.__version__) >= tokenize_version('4.2.1'):
        filter_key = 'filter_entries'
    return pyamg.smoothed_aggregation_solver(
        P, B=None,
        strength=('symmetric', {'theta': 0.0}),
        smooth=(
            'jacobi', {filter_key: True, 'weighting': 'local'}),
        improve_candidates=([('block_gauss_seidel',
                              {'sweep': 'symmetric', 'iterations': 4})] +
                            [None] * (max_levels - 1)),
        aggregate='standard',
        presmoother=('block_gauss_seidel',
                     {'sweep': 'symmetric', 'iterations': 1}),
        postsmoother=('block_gauss_seidel',
                      {'sweep': 'symmetric', 'iterations': 1}),
        max_levels=max_levels,
        max_coarse=300,
        coarse_solver='pinv')


THz = 1e12 * 1. / units.s


class Precon(ABC):

    @abstractmethod
    def make_precon(self, atoms, reinitialize=None):
        """
        Create a preconditioner matrix based on the passed set of atoms.

        Creates a general-purpose preconditioner for use with optimization
        algorithms, based on examining distances between pairs of atoms in the
        lattice. The matrix will be stored in the attribute self.P and
        returned.

        Args:
            atoms: the Atoms object used to create the preconditioner.

            reinitialize: if True, parameters of the preconditioner
                will be recalculated before the preconditioner matrix is
                created. If False, they will be calculated only when they
                do not currently have a value (ie, the first time this
                function is called).

        Returns:
            P: A sparse scipy csr_matrix. BE AWARE that using
                numpy.dot() with sparse matrices will result in
                errors/incorrect results - use the .dot method directly
                on the matrix instead.
        """
        ...

    @abstractmethod
    def Pdot(self, x):
        """
        Return the result of applying P to a vector x
        """
        ...

    def dot(self, x, y):
        """
        Return the preconditioned dot product <P x, y>

        Uses 128-bit floating point math for vector dot products
        """
        return longsum(self.Pdot(x) * y)

    def norm(self, x):
        """
        Return the P-norm of x, where |x|_P = sqrt(<Px, x>)
        """
        return np.sqrt(self.dot(x, x))

    def vdot(self, x, y):
        return self.dot(x.reshape(-1),
                        y.reshape(-1))

    @abstractmethod
    def solve(self, x):
        """
        Solve the (sparse) linear system P x = y and return y
        """
        ...

    def apply(self, forces, atoms):
        """
        Convenience wrapper that combines make_precon() and solve()

        Parameters
        ----------
        forces: array
            (len(atoms)*3) array of input forces
        atoms: ase.atoms.Atoms

        Returns
        -------
        precon_forces: array
            (len(atoms), 3) array of preconditioned forces
        residual: float
            inf-norm of original forces, i.e. maximum absolute force
        """
        self.make_precon(atoms)
        residual = np.linalg.norm(forces, np.inf)
        precon_forces = self.solve(forces)
        return precon_forces, residual

    @abstractmethod
    def copy(self):
        ...

    @abstractmethod
    def asarray(self):
        """
        Array representation of preconditioner, as a dense matrix
        """
        ...


class Logfile:
    def __init__(self, logfile=None):
        if isinstance(logfile, str):
            if logfile == "-":
                logfile = sys.stdout
            else:
                logfile = open(logfile, "a")
        self.logfile = logfile

    def write(self, *args):
        if self.logfile is None:
            return
        self.logfile.write(*args)


class SparsePrecon(Precon):
    def __init__(self, r_cut=None, r_NN=None,
                 mu=None, mu_c=None,
                 dim=3, c_stab=0.1, force_stab=False,
                 reinitialize=False, array_convention='C',
                 solver="auto", solve_tol=1e-8,
                 apply_positions=True, apply_cell=True,
                 estimate_mu_eigmode=False, logfile=None, rng=None,
                 neighbour_list=neighbor_list):
        """Initialise a preconditioner object based on passed parameters.

        Parameters:
            r_cut: float. This is a cut-off radius. The preconditioner matrix
                will be created by considering pairs of atoms that are within a
                distance r_cut of each other. For a regular lattice, this is
                usually taken somewhere between the first- and second-nearest
                neighbour distance. If r_cut is not provided, default is
                2 * r_NN (see below)
            r_NN: nearest neighbour distance. If not provided, this is
                  calculated
                from input structure.
            mu: float
                energy scale for position degreees of freedom. If `None`, mu
                is precomputed using finite difference derivatives.
            mu_c: float
                energy scale for cell degreees of freedom. Also precomputed
                if None.
            estimate_mu_eigmode:
                If True, estimates mu based on the lowest eigenmodes of
                unstabilised preconditioner. If False it uses the sine based
                approach.
            dim: int; dimensions of the problem
            c_stab: float. The diagonal of the preconditioner matrix will have
                a stabilisation constant added, which will be the value of
                c_stab times mu.
            force_stab:
                If True, always add the stabilisation to diagnonal, regardless
                of the presence of fixed atoms.
            reinitialize: if True, the value of mu will be recalculated when
                self.make_precon is called. This can be overridden in specific
                cases with reinitialize argument in self.make_precon. If it
                is set to True here, the value passed for mu will be
                irrelevant unless reinitialize is set to False the first time
                make_precon is called.
            array_convention: Either 'C' or 'F' for Fortran; this will change
                the preconditioner to reflect the ordering of the indices in
                the vector it will operate on. The C convention assumes the
                vector will be arranged atom-by-atom (ie [x1, y1, z1, x2, ...])
                while the F convention assumes it will be arranged component
                by component (ie [x1, x2, ..., y1, y2, ...]).
            solver: One of "auto", "direct" or "pyamg", specifying whether to
                use a direct sparse solver or PyAMG to solve P x = y.
                Default is "auto" which uses PyAMG if available, falling
                back to sparse solver if not. solve_tol: tolerance used for
                PyAMG sparse linear solver, if available.
            apply_positions:  bool
                if True, apply preconditioner to position DoF
            apply_cell: bool
                if True, apply preconditioner to cell DoF
            logfile: file object or str
                If *logfile* is a string, a file with that name will be opened.
                Use '-' for stdout.
            rng: None or np.random.RandomState instance
                Random number generator to use for initialising pyamg solver
            neighbor_list: function (optional). Optionally replace the built-in
                ASE neighbour list with an alternative with the same call
                signature, e.g. `matscipy.neighbours.neighbour_list`.

        Raises:
            ValueError for problem with arguments

        """
        self.r_NN = r_NN
        self.r_cut = r_cut
        self.mu = mu
        self.mu_c = mu_c
        self.estimate_mu_eigmode = estimate_mu_eigmode
        self.c_stab = c_stab
        self.force_stab = force_stab
        self.array_convention = array_convention
        self.reinitialize = reinitialize
        self.P = None
        self.old_positions = None

        use_pyamg = False
        if solver == "auto":
            use_pyamg = have_pyamg
        elif solver == "direct":
            use_pyamg = False
        elif solver == "pyamg":
            if not have_pyamg:
                raise RuntimeError("solver='pyamg', PyAMG can't be imported!")
            use_pyamg = True
        else:
            raise ValueError('unknown solver - '
                             'should be "auto", "direct" or "pyamg"')

        self.use_pyamg = use_pyamg
        self.solve_tol = solve_tol
        self.apply_positions = apply_positions
        self.apply_cell = apply_cell

        if dim < 1:
            raise ValueError('Dimension must be at least 1')
        self.dim = dim
        self.logfile = Logfile(logfile)

        if rng is None:
            rng = np.random.RandomState()
        self.rng = rng

        self.neighbor_list = neighbor_list

    def copy(self):
        return copy.deepcopy(self)

    def Pdot(self, x):
        return self.P.dot(x)

    def solve(self, x):
        start_time = time.time()
        if self.use_pyamg and have_pyamg:
            y = self.ml.solve(x, x0=self.rng.random(self.P.shape[0]),
                              tol=self.solve_tol,
                              accel='cg',
                              maxiter=300,
                              cycle='W')
        else:
            y = spsolve(self.P, x)
        self.logfile.write(
            f'--- Precon applied in {(time.time() - start_time)} seconds ---\n')
        return y

    def estimate_mu(self, atoms, H=None):
        r"""Estimate optimal preconditioner coefficient \mu

        \mu is estimated from a numerical solution of

            [dE(p+v) -  dE(p)] \cdot v = \mu < P1 v, v >

        with perturbation

            v(x,y,z) = H P_lowest_nonzero_eigvec(x, y, z)

            or

            v(x,y,z) = H (sin(x / Lx), sin(y / Ly), sin(z / Lz))

        After the optimal \mu is found, self.mu will be set to its value.

        If `atoms` is an instance of Filter an additional \mu_c
        will be computed for the cell degrees of freedom .

        Args:
            atoms: Atoms object for initial system

            H: 3x3 array or None
                Magnitude of deformation to apply.
                Default is 1e-2*rNN*np.eye(3)

        Returns:
            mu   : float
            mu_c : float or None
        """
        logfile = self.logfile

        if self.dim != 3:
            raise ValueError('Automatic calculation of mu only possible for '
                             'three-dimensional preconditioners. Try setting '
                             'mu manually instead.')

        if self.r_NN is None:
            self.r_NN = estimate_nearest_neighbour_distance(atoms,
                                                            self.neighbor_list)

        # deformation matrix, default is diagonal
        if H is None:
            H = 1e-2 * self.r_NN * np.eye(3)

        # compute perturbation
        p = atoms.get_positions()

        if self.estimate_mu_eigmode:
            self.mu = 1.0
            self.mu_c = 1.0
            c_stab = self.c_stab
            self.c_stab = 0.0

            if isinstance(atoms, Filter):
                n = len(atoms.atoms)
            else:
                n = len(atoms)
            self._make_sparse_precon(atoms, initial_assembly=True)
            self.P = self.P[:3 * n, :3 * n]
            eigvals, eigvecs = sparse.linalg.eigsh(self.P, k=4, which='SM')

            logfile.write('estimate_mu(): lowest 4 eigvals = %f %f %f %f\n' %
                          (eigvals[0], eigvals[1], eigvals[2], eigvals[3]))
            # check eigenvalues
            if any(eigvals[0:3] > 1e-6):
                raise ValueError('First 3 eigenvalues of preconditioner matrix'
                                 'do not correspond to translational modes.')
            elif eigvals[3] < 1e-6:
                raise ValueError('Fourth smallest eigenvalue of '
                                 'preconditioner matrix '
                                 'is too small, increase r_cut.')

            x = np.zeros(n)
            for i in range(n):
                x[i] = eigvecs[:, 3][3 * i]
            x = x / np.linalg.norm(x)
            if x[0] < 0:
                x = -x

            v = np.zeros(3 * len(atoms))
            for i in range(n):
                v[3 * i] = x[i]
                v[3 * i + 1] = x[i]
                v[3 * i + 2] = x[i]
            v = v / np.linalg.norm(v)
            v = v.reshape((-1, 3))

            self.c_stab = c_stab
        else:
            Lx, Ly, Lz = (p[:, i].max() - p[:, i].min() for i in range(3))
            logfile.write('estimate_mu(): Lx=%.1f Ly=%.1f Lz=%.1f\n' %
                          (Lx, Ly, Lz))

            x, y, z = p.T
            # sine_vr = [np.sin(x/Lx), np.sin(y/Ly), np.sin(z/Lz)], but we need
            # to take into account the possibility that one of Lx/Ly/Lz is
            # zero.
            sine_vr = [x, y, z]

            for i, L in enumerate([Lx, Ly, Lz]):
                if L == 0:
                    warnings.warn(
                        'Cell length L[%d] == 0. Setting H[%d,%d] = 0.' %
                        (i, i, i))
                    H[i, i] = 0.0
                else:
                    sine_vr[i] = np.sin(sine_vr[i] / L)

            v = np.dot(H, sine_vr).T

        natoms = len(atoms)
        if isinstance(atoms, Filter):
            natoms = len(atoms.atoms)
            eps = H / self.r_NN
            v[natoms:, :] = eps

        v1 = v.reshape(-1)

        # compute LHS
        dE_p = -atoms.get_forces().reshape(-1)
        atoms_v = atoms.copy()
        atoms_v.calc = atoms.calc
        if isinstance(atoms, Filter):
            atoms_v = atoms.__class__(atoms_v)
            if hasattr(atoms, 'constant_volume'):
                atoms_v.constant_volume = atoms.constant_volume
        atoms_v.set_positions(p + v)
        dE_p_plus_v = -atoms_v.get_forces().reshape(-1)

        # compute left hand side
        LHS = (dE_p_plus_v - dE_p) * v1

        # assemble P with \mu = 1
        self.mu = 1.0
        self.mu_c = 1.0

        self._make_sparse_precon(atoms, initial_assembly=True)

        # compute right hand side
        RHS = self.P.dot(v1) * v1

        # use partial sums to compute separate mu for positions and cell DoFs
        self.mu = longsum(LHS[:3 * natoms]) / longsum(RHS[:3 * natoms])
        if self.mu < 1.0:
            logfile.write('estimate_mu(): mu (%.3f) < 1.0, '
                          'capping at mu=1.0' % self.mu)
            self.mu = 1.0

        if isinstance(atoms, Filter):
            self.mu_c = longsum(LHS[3 * natoms:]) / longsum(RHS[3 * natoms:])
            if self.mu_c < 1.0:
                logfile.write('estimate_mu(): mu_c (%.3f) < 1.0, '
                              'capping at mu_c=1.0\n' % self.mu_c)
                self.mu_c = 1.0

        logfile.write('estimate_mu(): mu=%r, mu_c=%r\n' %
                      (self.mu, self.mu_c))

        self.P = None  # force a rebuild with new mu (there may be fixed atoms)
        return (self.mu, self.mu_c)

    def asarray(self):
        return np.array(self.P.todense())

    def one_dim_to_ndim(self, csc_P, N):
        """
        Expand an N x N precon matrix to self.dim*N x self.dim * N

        Args:
            csc_P (sparse matrix): N x N sparse matrix, in CSC format
        """
        start_time = time.time()
        if self.dim == 1:
            P = csc_P
        elif self.array_convention == 'F':
            csc_P = csc_P.tocsr()
            P = csc_P
            for _ in range(self.dim - 1):
                P = sparse.block_diag((P, csc_P)).tocsr()
        else:
            # convert back to triplet and read the arrays
            csc_P = csc_P.tocoo()
            i = csc_P.row * self.dim
            j = csc_P.col * self.dim
            z = csc_P.data

            # N-dimensionalise, interlaced coordinates
            I = np.hstack([i + d for d in range(self.dim)])
            J = np.hstack([j + d for d in range(self.dim)])
            Z = np.hstack([z for _ in range(self.dim)])
            P = sparse.csc_matrix((Z, (I, J)),
                                  shape=(self.dim * N, self.dim * N))
            P = P.tocsr()
        self.logfile.write(
            f'--- N-dim precon created in {(time.time() - start_time)} s ---\n')
        return P

    def create_solver(self):
        if self.use_pyamg and have_pyamg:
            start_time = time.time()
            self.ml = create_pyamg_solver(self.P)
            self.logfile.write(
                f'--- multi grid solver created in {(time.time() - start_time)}'
                ' ---\n')


class SparseCoeffPrecon(SparsePrecon):
    def _make_sparse_precon(self, atoms, initial_assembly=False,
                            force_stab=False):
        """Create a sparse preconditioner matrix based on the passed atoms.

        Creates a general-purpose preconditioner for use with optimization
        algorithms, based on examining distances between pairs of atoms in the
        lattice. The matrix will be stored in the attribute self.P and
        returned. Note that this function will use self.mu, whatever it is.

        Args:
            atoms: the Atoms object used to create the preconditioner.

        Returns:
            A scipy.sparse.csr_matrix object, representing a d*N by d*N matrix
            (where N is the number of atoms, and d is the value of self.dim).
            BE AWARE that using numpy.dot() with this object will result in
            errors/incorrect results - use the .dot method directly on the
            sparse matrix instead.

        """
        logfile = self.logfile
        logfile.write('creating sparse precon: initial_assembly=%r, '
                      'force_stab=%r, apply_positions=%r, apply_cell=%r\n' %
                      (initial_assembly, force_stab, self.apply_positions,
                       self.apply_cell))

        N = len(atoms)
        diag_i = np.arange(N, dtype=int)
        start_time = time.time()
        if self.apply_positions:
            # compute neighbour list
            i, j, rij, fixed_atoms = get_neighbours(
                atoms, self.r_cut,
                neighbor_list=self.neighbor_list)
            logfile.write(
                f'--- neighbour list created in {(time.time() - start_time)} s '
                '--- \n')

            # compute entries in triplet format: without the constraints
            start_time = time.time()
            coeff = self.get_coeff(rij)
            diag_coeff = np.bincount(i, -coeff, minlength=N).astype(np.float64)
            if force_stab or len(fixed_atoms) == 0:
                logfile.write('adding stabilisation to precon')
                diag_coeff += self.mu * self.c_stab
        else:
            diag_coeff = np.ones(N)

        # precon is mu_c * identity for cell DoF
        if isinstance(atoms, Filter):
            if self.apply_cell:
                diag_coeff[-3:] = self.mu_c
            else:
                diag_coeff[-3:] = 1.0
        logfile.write(
            f'--- computed triplet format in {(time.time() - start_time)} s '
            '---\n')

        if self.apply_positions and not initial_assembly:
            # apply the constraints
            start_time = time.time()
            mask = np.ones(N)
            mask[fixed_atoms] = 0.0
            coeff *= mask[i] * mask[j]
            diag_coeff[fixed_atoms] = 1.0
            logfile.write(
                f'--- applied fixed_atoms in {(time.time() - start_time)} s '
                '---\n')

        if self.apply_positions:
            # remove zeros
            start_time = time.time()
            inz = np.nonzero(coeff)
            i = np.hstack((i[inz], diag_i))
            j = np.hstack((j[inz], diag_i))
            coeff = np.hstack((coeff[inz], diag_coeff))
            logfile.write(
                f'--- remove zeros in {(time.time() - start_time)} s '
                '---\n')
        else:
            i = diag_i
            j = diag_i
            coeff = diag_coeff

        # create an N x N precon matrix in compressed sparse column (CSC) format
        start_time = time.time()
        csc_P = sparse.csc_matrix((coeff, (i, j)), shape=(N, N))
        logfile.write(
            f'--- created CSC matrix in {(time.time() - start_time)} s '
            '---\n')

        self.P = self.one_dim_to_ndim(csc_P, N)
        self.create_solver()

    def make_precon(self, atoms, reinitialize=None):
        if self.r_NN is None:
            self.r_NN = estimate_nearest_neighbour_distance(atoms,
                                                            self.neighbor_list)

        if self.r_cut is None:
            # This is the first time this function has been called, and no
            # cutoff radius has been specified, so calculate it automatically.
            self.r_cut = 2.0 * self.r_NN
        elif self.r_cut < self.r_NN:
            warning = ('WARNING: r_cut (%.2f) < r_NN (%.2f), '
                       'increasing to 1.1*r_NN = %.2f' % (self.r_cut,
                                                          self.r_NN,
                                                          1.1 * self.r_NN))
            warnings.warn(warning)
            self.r_cut = 1.1 * self.r_NN

        if reinitialize is None:
            # The caller has not specified whether or not to recalculate mu,
            # so the Precon's setting is used.
            reinitialize = self.reinitialize

        if self.mu is None:
            # Regardless of what the caller has specified, if we don't
            # currently have a value of mu, then we need one.
            reinitialize = True

        if reinitialize:
            self.estimate_mu(atoms)

        if self.P is not None:
            real_atoms = atoms
            if isinstance(atoms, Filter):
                real_atoms = atoms.atoms
            if self.old_positions is None:
                self.old_positions = real_atoms.positions
            displacement, _ = find_mic(real_atoms.positions -
                                       self.old_positions,
                                       real_atoms.cell, real_atoms.pbc)
            self.old_positions = real_atoms.get_positions()
            max_abs_displacement = abs(displacement).max()
            self.logfile.write('max(abs(displacements)) = %.2f A (%.2f r_NN)' %
                               (max_abs_displacement,
                                max_abs_displacement / self.r_NN))
            if max_abs_displacement < 0.5 * self.r_NN:
                return

        start_time = time.time()
        self._make_sparse_precon(atoms, force_stab=self.force_stab)
        self.logfile.write(
            f'--- Precon created in {(time.time() - start_time)} seconds '
            '--- \n')

    @abstractmethod
    def get_coeff(self, r):
        ...


class Pfrommer(Precon):
    """
    Use initial guess for inverse Hessian from Pfrommer et al. as a
    simple preconditioner

    J. Comput. Phys. vol 131 p233-240 (1997)
    """

    def __init__(self, bulk_modulus=500 * units.GPa, phonon_frequency=50 * THz,
                 apply_positions=True, apply_cell=True):
        """
        Default bulk modulus is 500 GPa and default phonon frequency is 50 THz
        """
        self.bulk_modulus = bulk_modulus
        self.phonon_frequency = phonon_frequency
        self.apply_positions = apply_positions
        self.apply_cell = apply_cell
        self.H0 = None

    def make_precon(self, atoms, reinitialize=None):
        if self.H0 is not None:
            # only build H0 on first call
            return

        variable_cell = False
        if isinstance(atoms, Filter):
            variable_cell = True
            atoms = atoms.atoms

        # position DoF
        omega = self.phonon_frequency
        mass = atoms.get_masses().mean()
        block = np.eye(3) / (mass * omega**2)
        blocks = [block] * len(atoms)

        # cell DoF
        if variable_cell:
            coeff = 1.0
            if self.apply_cell:
                coeff = 1.0 / (3 * self.bulk_modulus)
            blocks.append(np.diag([coeff] * 9))

        self.H0 = sparse.block_diag(blocks, format='csr')
        return

    def Pdot(self, x):
        return self.H0.solve(x)

    def solve(self, x):
        y = self.H0.dot(x)
        return y

    def copy(self):
        return Pfrommer(self.bulk_modulus,
                        self.phonon_frequency,
                        self.apply_positions,
                        self.apply_cell)

    def asarray(self):
        return np.array(self.H0.todense())


class IdentityPrecon(Precon):
    """
    Dummy preconditioner which does not modify forces
    """

    def make_precon(self, atoms, reinitialize=None):
        self.atoms = atoms

    def Pdot(self, x):
        return x

    def solve(self, x):
        return x

    def copy(self):
        return IdentityPrecon()

    def asarray(self):
        return np.eye(3 * len(self.atoms))


class C1(SparseCoeffPrecon):
    """Creates matrix by inserting a constant whenever r_ij is less than r_cut.
    """

    def __init__(self, r_cut=None, mu=None, mu_c=None, dim=3, c_stab=0.1,
                 force_stab=False,
                 reinitialize=False, array_convention='C',
                 solver="auto", solve_tol=1e-9,
                 apply_positions=True, apply_cell=True, logfile=None):
        super().__init__(r_cut=r_cut, mu=mu, mu_c=mu_c,
                         dim=dim, c_stab=c_stab,
                         force_stab=force_stab,
                         reinitialize=reinitialize,
                         array_convention=array_convention,
                         solver=solver, solve_tol=solve_tol,
                         apply_positions=apply_positions,
                         apply_cell=apply_cell,
                         logfile=logfile)

    def get_coeff(self, r):
        return -self.mu * np.ones_like(r)


class Exp(SparseCoeffPrecon):
    """Creates matrix with values decreasing exponentially with distance.
    """

    def __init__(self, A=3.0, r_cut=None, r_NN=None, mu=None, mu_c=None,
                 dim=3, c_stab=0.1,
                 force_stab=False, reinitialize=False, array_convention='C',
                 solver="auto", solve_tol=1e-9,
                 apply_positions=True, apply_cell=True,
                 estimate_mu_eigmode=False, logfile=None):
        """
        Initialise an Exp preconditioner with given parameters.

        Args:
            r_cut, mu, c_stab, dim, sparse, reinitialize, array_convention: see
                precon.__init__()
            A: coefficient in exp(-A*r/r_NN). Default is A=3.0.
        """
        super().__init__(r_cut=r_cut, r_NN=r_NN,
                         mu=mu, mu_c=mu_c, dim=dim, c_stab=c_stab,
                         force_stab=force_stab,
                         reinitialize=reinitialize,
                         array_convention=array_convention,
                         solver=solver,
                         solve_tol=solve_tol,
                         apply_positions=apply_positions,
                         apply_cell=apply_cell,
                         estimate_mu_eigmode=estimate_mu_eigmode,
                         logfile=logfile)

        self.A = A

    def get_coeff(self, r):
        return -self.mu * np.exp(-self.A * (r / self.r_NN - 1))


def ij_to_x(i, j):
    x = [3 * i, 3 * i + 1, 3 * i + 2,
         3 * j, 3 * j + 1, 3 * j + 2]
    return x


def ijk_to_x(i, j, k):
    x = [3 * i, 3 * i + 1, 3 * i + 2,
         3 * j, 3 * j + 1, 3 * j + 2,
         3 * k, 3 * k + 1, 3 * k + 2]
    return x


def ijkl_to_x(i, j, k, l):
    x = [3 * i, 3 * i + 1, 3 * i + 2,
         3 * j, 3 * j + 1, 3 * j + 2,
         3 * k, 3 * k + 1, 3 * k + 2,
         3 * l, 3 * l + 1, 3 * l + 2]
    return x


def apply_fixed(atoms, P):
    fixed_atoms = []
    for constraint in atoms.constraints:
        if isinstance(constraint, FixAtoms):
            fixed_atoms.extend(list(constraint.index))
        else:
            raise TypeError(
                'only FixAtoms constraints are supported by Precon class')
    if len(fixed_atoms) != 0:
        P = P.tolil()
    for i in fixed_atoms:
        P[i, :] = 0.0
        P[:, i] = 0.0
        P[i, i] = 1.0
    return P


class FF(SparsePrecon):
    """Creates matrix using morse/bond/angle/dihedral force field parameters.
    """

    def __init__(self, dim=3, c_stab=0.1, force_stab=False,
                 array_convention='C', solver="auto", solve_tol=1e-9,
                 apply_positions=True, apply_cell=True,
                 hessian='spectral', morses=None, bonds=None, angles=None,
                 dihedrals=None, logfile=None):
        """Initialise an FF preconditioner with given parameters.

        Args:
             dim, c_stab, force_stab, array_convention, use_pyamg, solve_tol:
                see SparsePrecon.__init__()
             morses: Morse instance
             bonds: Bond instance
             angles: Angle instance
             dihedrals: Dihedral instance
        """

        if (morses is None and bonds is None and angles is None and
                dihedrals is None):
            raise ImportError(
                'At least one of morses, bonds, angles or dihedrals must be '
                'defined!')

        super().__init__(dim=dim, c_stab=c_stab,
                         force_stab=force_stab,
                         array_convention=array_convention,
                         solver=solver,
                         solve_tol=solve_tol,
                         apply_positions=apply_positions,
                         apply_cell=apply_cell, logfile=logfile)

        self.hessian = hessian
        self.morses = morses
        self.bonds = bonds
        self.angles = angles
        self.dihedrals = dihedrals

    def make_precon(self, atoms, reinitialize=None):
        start_time = time.time()
        self._make_sparse_precon(atoms, force_stab=self.force_stab)
        self.logfile.write(
            f'--- Precon created in {(time.time() - start_time)} seconds '
            '---\n')

    def add_morse(self, morse, atoms, row, col, data, conn=None):
        if self.hessian == 'reduced':
            i, j, Hx = ff.get_morse_potential_reduced_hessian(
                atoms, morse)
        elif self.hessian == 'spectral':
            i, j, Hx = ff.get_morse_potential_hessian(
                atoms, morse, spectral=True)
        else:
            raise NotImplementedError('Not implemented hessian')
        x = ij_to_x(i, j)
        row.extend(np.repeat(x, 6))
        col.extend(np.tile(x, 6))
        data.extend(Hx.flatten())
        if conn is not None:
            conn[i, j] = True
            conn[j, i] = True

    def add_bond(self, bond, atoms, row, col, data, conn=None):
        if self.hessian == 'reduced':
            i, j, Hx = ff.get_bond_potential_reduced_hessian(
                atoms, bond, self.morses)
        elif self.hessian == 'spectral':
            i, j, Hx = ff.get_bond_potential_hessian(
                atoms, bond, self.morses, spectral=True)
        else:
            raise NotImplementedError('Not implemented hessian')
        x = ij_to_x(i, j)
        row.extend(np.repeat(x, 6))
        col.extend(np.tile(x, 6))
        data.extend(Hx.flatten())
        if conn is not None:
            conn[i, j] = True
            conn[j, i] = True

    def add_angle(self, angle, atoms, row, col, data, conn=None):
        if self.hessian == 'reduced':
            i, j, k, Hx = ff.get_angle_potential_reduced_hessian(
                atoms, angle, self.morses)
        elif self.hessian == 'spectral':
            i, j, k, Hx = ff.get_angle_potential_hessian(
                atoms, angle, self.morses, spectral=True)
        else:
            raise NotImplementedError('Not implemented hessian')
        x = ijk_to_x(i, j, k)
        row.extend(np.repeat(x, 9))
        col.extend(np.tile(x, 9))
        data.extend(Hx.flatten())
        if conn is not None:
            conn[i, j] = conn[i, k] = conn[j, k] = True
            conn[j, i] = conn[k, i] = conn[k, j] = True

    def add_dihedral(self, dihedral, atoms, row, col, data, conn=None):
        if self.hessian == 'reduced':
            i, j, k, l, Hx = \
                ff.get_dihedral_potential_reduced_hessian(
                    atoms, dihedral, self.morses)
        elif self.hessian == 'spectral':
            i, j, k, l, Hx = ff.get_dihedral_potential_hessian(
                atoms, dihedral, self.morses, spectral=True)
        else:
            raise NotImplementedError('Not implemented hessian')
        x = ijkl_to_x(i, j, k, l)
        row.extend(np.repeat(x, 12))
        col.extend(np.tile(x, 12))
        data.extend(Hx.flatten())
        if conn is not None:
            conn[i, j] = conn[i, k] = conn[i, l] = conn[
                j, k] = conn[j, l] = conn[k, l] = True
            conn[j, i] = conn[k, i] = conn[l, i] = conn[
                k, j] = conn[l, j] = conn[l, k] = True

    def _make_sparse_precon(self, atoms, initial_assembly=False,
                            force_stab=False):
        N = len(atoms)

        row = []
        col = []
        data = []

        if self.morses is not None:
            for morse in self.morses:
                self.add_morse(morse, atoms, row, col, data)

        if self.bonds is not None:
            for bond in self.bonds:
                self.add_bond(bond, atoms, row, col, data)

        if self.angles is not None:
            for angle in self.angles:
                self.add_angle(angle, atoms, row, col, data)

        if self.dihedrals is not None:
            for dihedral in self.dihedrals:
                self.add_dihedral(dihedral, atoms, row, col, data)

        # add the diagonal
        row.extend(range(self.dim * N))
        col.extend(range(self.dim * N))
        data.extend([self.c_stab] * self.dim * N)

        # create the matrix
        start_time = time.time()
        self.P = sparse.csc_matrix(
            (data, (row, col)), shape=(self.dim * N, self.dim * N))
        self.logfile.write(
            f'--- created CSC matrix in {(time.time() - start_time)} s ---\n')

        self.P = apply_fixed(atoms, self.P)
        self.P = self.P.tocsr()
        self.logfile.write(
            f'--- N-dim precon created in {(time.time() - start_time)} s ---\n')
        self.create_solver()


class Exp_FF(Exp, FF):
    """Creates matrix with values decreasing exponentially with distance.
    """

    def __init__(self, A=3.0, r_cut=None, r_NN=None, mu=None, mu_c=None,
                 dim=3, c_stab=0.1,
                 force_stab=False, reinitialize=False, array_convention='C',
                 solver="auto", solve_tol=1e-9,
                 apply_positions=True, apply_cell=True,
                 estimate_mu_eigmode=False,
                 hessian='spectral', morses=None, bonds=None, angles=None,
                 dihedrals=None, logfile=None):
        """Initialise an Exp+FF preconditioner with given parameters.

        Args:
            r_cut, mu, c_stab, dim, reinitialize, array_convention: see
                precon.__init__()
            A: coefficient in exp(-A*r/r_NN). Default is A=3.0.
        """
        if (morses is None and bonds is None and angles is None and
                dihedrals is None):
            raise ImportError(
                'At least one of morses, bonds, angles or dihedrals must '
                'be defined!')

        SparsePrecon.__init__(self, r_cut=r_cut, r_NN=r_NN,
                              mu=mu, mu_c=mu_c, dim=dim, c_stab=c_stab,
                              force_stab=force_stab,
                              reinitialize=reinitialize,
                              array_convention=array_convention,
                              solver=solver,
                              solve_tol=solve_tol,
                              apply_positions=apply_positions,
                              apply_cell=apply_cell,
                              estimate_mu_eigmode=estimate_mu_eigmode,
                              logfile=logfile)

        self.A = A
        self.hessian = hessian
        self.morses = morses
        self.bonds = bonds
        self.angles = angles
        self.dihedrals = dihedrals

    def make_precon(self, atoms, reinitialize=None):
        if self.r_NN is None:
            self.r_NN = estimate_nearest_neighbour_distance(atoms,
                                                            self.neighbor_list)

        if self.r_cut is None:
            # This is the first time this function has been called, and no
            # cutoff radius has been specified, so calculate it automatically.
            self.r_cut = 2.0 * self.r_NN
        elif self.r_cut < self.r_NN:
            warning = ('WARNING: r_cut (%.2f) < r_NN (%.2f), '
                       'increasing to 1.1*r_NN = %.2f' % (self.r_cut,
                                                          self.r_NN,
                                                          1.1 * self.r_NN))
            warnings.warn(warning)
            self.r_cut = 1.1 * self.r_NN

        if reinitialize is None:
            # The caller has not specified whether or not to recalculate mu,
            # so the Precon's setting is used.
            reinitialize = self.reinitialize

        if self.mu is None:
            # Regardless of what the caller has specified, if we don't
            # currently have a value of mu, then we need one.
            reinitialize = True

        if reinitialize:
            self.estimate_mu(atoms)

        if self.P is not None:
            real_atoms = atoms
            if isinstance(atoms, Filter):
                real_atoms = atoms.atoms
            if self.old_positions is None:
                self.old_positions = real_atoms.positions
            displacement, _ = find_mic(real_atoms.positions -
                                       self.old_positions,
                                       real_atoms.cell, real_atoms.pbc)
            self.old_positions = real_atoms.get_positions()
            max_abs_displacement = abs(displacement).max()
            self.logfile.write('max(abs(displacements)) = %.2f A (%.2f r_NN)' %
                               (max_abs_displacement,
                                max_abs_displacement / self.r_NN))
            if max_abs_displacement < 0.5 * self.r_NN:
                return

        # Create the preconditioner:
        start_time = time.time()
        self._make_sparse_precon(atoms, force_stab=self.force_stab)
        self.logfile.write(
            f'--- Precon created in {(time.time() - start_time)} seconds ---\n')

    def _make_sparse_precon(self, atoms, initial_assembly=False,
                            force_stab=False):
        """Create a sparse preconditioner matrix based on the passed atoms.

        Args:
            atoms: the Atoms object used to create the preconditioner.

        Returns:
            A scipy.sparse.csr_matrix object, representing a d*N by d*N matrix
            (where N is the number of atoms, and d is the value of self.dim).
            BE AWARE that using numpy.dot() with this object will result in
            errors/incorrect results - use the .dot method directly on the
            sparse matrix instead.

        """
        self.logfile.write('creating sparse precon: initial_assembly=%r, '
                           'force_stab=%r, apply_positions=%r, '
                           'apply_cell=%r\n' %
                           (initial_assembly, force_stab,
                            self.apply_positions, self.apply_cell))

        N = len(atoms)
        start_time = time.time()
        if self.apply_positions:
            # compute neighbour list
            i_list, j_list, rij_list, _fixed_atoms = get_neighbours(
                atoms, self.r_cut, self.neighbor_list)
            self.logfile.write(
                f'--- neighbour list created in {(time.time() - start_time)} s '
                '---\n')

        row = []
        col = []
        data = []

        # precon is mu_c*identity for cell DoF
        start_time = time.time()
        if isinstance(atoms, Filter):
            i = N - 3
            j = N - 2
            k = N - 1
            x = ijk_to_x(i, j, k)
            row.extend(x)
            col.extend(x)
            if self.apply_cell:
                data.extend(np.repeat(self.mu_c, 9))
            else:
                data.extend(np.repeat(self.mu_c, 9))
        self.logfile.write(
            f'--- computed triplet format in {(time.time() - start_time)} s '
            '---\n')

        conn = sparse.lil_matrix((N, N), dtype=bool)

        if self.apply_positions and not initial_assembly:
            if self.morses is not None:
                for morse in self.morses:
                    self.add_morse(morse, atoms, row, col, data, conn)

            if self.bonds is not None:
                for bond in self.bonds:
                    self.add_bond(bond, atoms, row, col, data, conn)

            if self.angles is not None:
                for angle in self.angles:
                    self.add_angle(angle, atoms, row, col, data, conn)

            if self.dihedrals is not None:
                for dihedral in self.dihedrals:
                    self.add_dihedral(dihedral, atoms, row, col, data, conn)

        if self.apply_positions:
            for i, j, rij in zip(i_list, j_list, rij_list):
                if not conn[i, j]:
                    coeff = self.get_coeff(rij)
                    x = ij_to_x(i, j)
                    row.extend(x)
                    col.extend(x)
                    data.extend(3 * [-coeff] + 3 * [coeff])

        row.extend(range(self.dim * N))
        col.extend(range(self.dim * N))
        if initial_assembly:
            data.extend([self.mu * self.c_stab] * self.dim * N)
        else:
            data.extend([self.c_stab] * self.dim * N)

        # create the matrix
        start_time = time.time()
        self.P = sparse.csc_matrix(
            (data, (row, col)), shape=(self.dim * N, self.dim * N))
        self.logfile.write(
            f'--- created CSC matrix in {(time.time() - start_time)} s ---\n')

        if not initial_assembly:
            self.P = apply_fixed(atoms, self.P)

        self.P = self.P.tocsr()
        self.create_solver()


def make_precon(precon, atoms=None, **kwargs):
    """
    Construct preconditioner from a string and optionally build for Atoms

    Parameters
    ----------
    precon - one of 'C1', 'Exp', 'Pfrommer', 'FF', 'Exp_FF', 'ID', None
             or an instance of a subclass of `ase.optimize.precon.Precon`

    atoms - ase.atoms.Atoms instance, optional
            If present, build apreconditioner for this Atoms object

    **kwargs - additional keyword arguments to pass to Precon constructor

    Returns
    -------
    precon - instance of relevant subclass of `ase.optimize.precon.Precon`
    """
    lookup = {
        'C1': C1,
        'Exp': Exp,
        'Pfrommer': Pfrommer,
        'FF': FF,
        'Exp_FF': Exp_FF,
        'ID': IdentityPrecon,
        'IdentityPrecon': IdentityPrecon,
        None: IdentityPrecon
    }
    if isinstance(precon, str) or precon is None:
        cls = lookup[precon]
        precon = cls(**kwargs)
    if atoms is not None:
        precon.make_precon(atoms)
    return precon


class SplineFit:
    """
    Fit a cubic spline fit to images
    """

    def __init__(self, s, x):
        self._s = s
        self._x_data = x
        self._x = CubicSpline(self._s, x, bc_type='not-a-knot')
        self._dx_ds = self._x.derivative()
        self._d2x_ds2 = self._x.derivative(2)

    @property
    def s(self):
        return self._s

    @property
    def x_data(self):
        return self._x_data

    @property
    def x(self):
        return self._x

    @property
    def dx_ds(self):
        return self._dx_ds

    @property
    def d2x_ds2(self):
        return self._d2x_ds2


class PreconImages:
    def __init__(self, precon, images, **kwargs):
        """
        Wrapper for a list of Precon objects and associated images

        This is used when preconditioning a NEB object. Equation references
        refer to Paper IV in the :class:`ase.mep.NEB` documentation, i.e.

        S. Makri, C. Ortner and J. R. Kermode, J. Chem. Phys.
        150, 094109 (2019)
        https://dx.doi.org/10.1063/1.5064465

        Args:
            precon (str or list): preconditioner(s) to use
            images (list of Atoms): Atoms objects that define the state

        """
        self.images = images
        self._spline = None

        if isinstance(precon, list):
            if len(precon) != len(images):
                raise ValueError(f'length mismatch: len(precon)={len(precon)} '
                                 f'!= len(images)={len(images)}')
            self.precon = precon
            return
        P0 = make_precon(precon, images[0], **kwargs)
        self.precon = [P0]
        for image in images[1:]:
            P = P0.copy()
            P.make_precon(image)
            self.precon.append(P)

    def __len__(self):
        return len(self.precon)

    def __iter__(self):
        return iter(self.precon)

    def __getitem__(self, index):
        return self.precon[index]

    def apply(self, all_forces, index=None):
        """Apply preconditioners to stored images

        Args:
            all_forces (array): forces on images, shape (nimages, natoms, 3)
            index (slice, optional): Which images to include. Defaults to all.

        Returns:
            precon_forces: array of preconditioned forces
        """
        if index is None:
            index = slice(None)
        precon_forces = []
        for precon, image, forces in zip(self.precon[index],
                                         self.images[index],
                                         all_forces):
            f_vec = forces.reshape(-1)
            pf_vec, _ = precon.apply(f_vec, image)
            precon_forces.append(pf_vec.reshape(-1, 3))

        return np.array(precon_forces)

    def average_norm(self, i, j, dx):
        """Average norm between images i and j

        Args:
            i (int): left image
            j (int): right image
            dx (array): vector

        Returns:
            norm: norm of vector wrt average of precons at i and j
        """
        return np.sqrt(0.5 * (self.precon[i].dot(dx, dx) +
                              self.precon[j].dot(dx, dx)))

    def get_tangent(self, i):
        """Normalised tangent vector at image i

        Args:
            i (int): image of interest

        Returns:
            tangent: tangent vector, normalised with appropriate precon norm
        """
        tangent = self.spline.dx_ds(self.spline.s[i])
        tangent /= self.precon[i].norm(tangent)
        return tangent.reshape(-1, 3)

    def get_residual(self, i, imgforce):
        # residuals computed according to eq. 11 in the paper
        P_dot_imgforce = self.precon[i].Pdot(imgforce.reshape(-1))
        return np.linalg.norm(P_dot_imgforce, np.inf)

    def get_spring_force(self, i, k1, k2, tangent):
        """Spring force on image

        Args:
            i (int): image of interest
            k1 (float): spring constant for left spring
            k2 (float): spring constant for right spring
            tangent (array): tangent vector, shape (natoms, 3)

        Returns:
            eta: NEB spring forces, shape (natoms, 3)
        """
        # Definition following Eq. 9 in Paper IV
        nimages = len(self.images)
        k = 0.5 * (k1 + k2) / (nimages ** 2)
        curvature = self.spline.d2x_ds2(self.spline.s[i]).reshape(-1, 3)
        # complete Eq. 9 by including the spring force
        eta = k * self.precon[i].vdot(curvature, tangent) * tangent
        return eta

    def get_coordinates(self, positions=None):
        """Compute displacements wrt appropriate precon metric for each image

        Args:
            positions (list or array, optional) - images positions.
                Shape either (nimages * natoms, 3) or ((nimages-2)*natoms, 3)

        Returns:
            s : array shape (nimages,), reaction coordinates, in range [0, 1]
            x : array shape (nimages, 3 * natoms), flat displacement vectors
        """
        nimages = len(self.precon)
        natoms = len(self.images[0])
        d_P = np.zeros(nimages)
        x = np.zeros((nimages, 3 * natoms))  # flattened positions
        if positions is None:
            positions = [image.positions for image in self.images]
        elif isinstance(positions, np.ndarray) and len(positions.shape) == 2:
            positions = positions.reshape(-1, natoms, 3)
            positions = [positions[i, :, :] for i in range(len(positions))]
            if len(positions) == len(self.images) - 2:
                # prepend and append the non-moving images
                positions = ([self.images[0].positions] + positions +
                             [self.images[-1].positions])
        assert len(positions) == len(self.images)

        x[0, :] = positions[0].reshape(-1)
        for i in range(1, nimages):
            x[i, :] = positions[i].reshape(-1)
            dx, _ = find_mic(positions[i] - positions[i - 1],
                             self.images[i - 1].cell,
                             self.images[i - 1].pbc)
            dx = dx.reshape(-1)
            d_P[i] = self.average_norm(i, i - 1, dx)

        s = d_P.cumsum() / d_P.sum()  # Eq. A1 in paper IV
        return s, x

    def spline_fit(self, positions=None):
        """Fit 3 * natoms cubic splines as a function of reaction coordinate

        Returns:
            fit : :class:`ase.optimize.precon.SplineFit` object
        """
        s, x = self.get_coordinates(positions)
        return SplineFit(s, x)

    @property
    def spline(self):
        s, x = self.get_coordinates()
        if self._spline and (np.abs(s - self._old_s).max() < 1e-6 and
                             np.abs(x - self._old_x).max() < 1e-6):
            return self._spline

        self._spline = self.spline_fit()
        self._old_s = s
        self._old_x = x
        return self._spline



from ase.build import stack
from ase.build.surface import _all_surface_functions
from ase.calculators.calculator import compare_atoms


def test_surface_stack():

    # The purpose of this test is to test the stack() function and verify
    # that the various surface builder functions produce configurations
    # consistent with stacking.

    d = _all_surface_functions()
    exclude = {'mx2', 'graphene'}  # mx2 and graphene are not like the others

    for name in sorted(d):
        if name in exclude:
            continue

        func = d[name]

        def has(var):
            c = func.__code__
            return var in c.co_varnames[:c.co_argcount]

        for nlayers in range(1, 7):
            atoms = func('Au', size=(2, 2, nlayers), periodic=True, a=4.0)
            big_atoms = func('Au', size=(2, 2, 2 * nlayers),
                             periodic=True, a=4.0)
            stacked_atoms = stack(atoms, atoms)

            changes = compare_atoms(stacked_atoms, big_atoms, tol=1e-11)
            if not changes:
                print('OK', name, nlayers)
                break
        else:
            assert 0, f'Unstackable surface {name}'



import numpy as np
import pytest

from ase.build import fcc111, molecule
from ase.build.attach import (
    attach,
    attach_randomly,
    attach_randomly_and_broadcast,
)
from ase.parallel import world


def test_attach_molecules():
    """Attach two molecules and check that their minimal distance
    is as required"""
    m1 = molecule('C6H6')
    m2 = molecule('NH3')

    distance = 2.
    m12 = attach(m1, m2, distance)
    dmin = np.linalg.norm(m12[15].position - m12[8].position)
    assert dmin == pytest.approx(distance, 1e-8)


def test_pbc():
    """Attach two molecules and check attachment considers pbc"""
    m1 = molecule('C6H6')
    m1.cell = (20, 1, 1)
    m1.translate((16, 0, 0))
    m1.pbc = (1, 0, 0)
    m2 = molecule('NH3')

    distance = 2.
    m12 = attach(m1, m2, distance)
    for atom in m12[-4:]:
        assert atom.position[0] < 2


def test_attach_to_surface():
    """Attach a molecule to a surafce at a given distance"""
    slab = fcc111('Al', size=(3, 2, 2), vacuum=10.0)
    mol = molecule('CH4')

    distance = 3.
    struct = attach(slab, mol, distance, (0, 0, 1))
    dmin = np.linalg.norm(struct[6].position - struct[15].position)
    assert dmin == pytest.approx(distance, 1e-8)


def test_attach_randomly():
    """Attach two molecules in random orientation."""
    m1 = molecule('C6H6')
    m2 = molecule('CF4')
    distance = 2.5

    if world.size > 1:
        "Check that the coordinates are correctly distributed from master."
        rng = np.random.RandomState(world.rank)  # ensure different seed
        atoms = attach_randomly_and_broadcast(m1, m2, distance, rng)

        p0 = 1. * atoms[-1].position
        world.broadcast(p0, 0)
        for i in range(1, world.size):
            pi = 1. * atoms[-1].position
            world.broadcast(pi, i)
            assert pi == pytest.approx(p0, 1e-8)

        "Check that every core has its own structure"
        rng = np.random.RandomState(world.rank)  # ensure different seed
        atoms = attach_randomly(m1, m2, distance, rng)
        p0 = 1. * atoms[-1].position
        world.broadcast(p0, 0)
        for i in range(1, world.size):
            pi = 1. * atoms[-1].position
            world.broadcast(pi, i)
            assert pi != pytest.approx(p0, 1e-8)

    rng = np.random.RandomState(42)  # ensure the same seed
    pos2_ac = np.zeros((5, 3))
    N = 25
    for _ in range(N):
        atoms = attach_randomly(m1, m2, distance, rng=rng)
        pos2_ac += atoms.get_positions()[12:, :]
    # the average position should be "zero" approximately
    assert (np.abs(pos2_ac / N) <= 1).all()



import numpy as np
import pytest

from ase.build import bulk
from ase.calculators.test import FreeElectrons
from ase.dft.kpoints import special_paths
from ase.lattice import RHL
from ase.spectrum.band_structure import BandStructure


def test_bandstructure(testdir, plt):
    atoms = bulk('Cu')
    path = special_paths['fcc']
    atoms.calc = FreeElectrons(nvalence=1,
                               kpts={'path': path, 'npoints': 200})
    atoms.get_potential_energy()
    bs = atoms.calc.band_structure()
    _coords, _labelcoords, labels = bs.get_labels()
    print(labels)
    bs.write('hmm.json')
    bs = BandStructure.read('hmm.json')
    _coords, _labelcoords, labels = bs.get_labels()
    print(labels)
    assert ''.join(labels) == 'GXWKGLUWLKUX'
    bs.plot(emax=10, filename='bs.png')
    cols = np.linspace(-1.0, 1.0, bs.energies.size)
    cols.shape = bs.energies.shape
    bs.plot(emax=10, point_colors=cols, filename='bs2.png')


@pytest.fixture()
def bs():
    rhl = RHL(4.0, 65.0)
    path = rhl.bandpath()
    return path.free_electron_band_structure()


def test_print_bs(bs):
    print(bs)


def test_subtract_ref(bs):
    avg = np.mean(bs.energies)
    bs._reference = 5
    bs2 = bs.subtract_reference()
    avg2 = np.mean(bs2.energies)
    assert avg - 5 == pytest.approx(avg2)



import sys
from math import factorial

import numpy as np
import pytest

from ase.build import molecule
from ase.calculators.emt import EMT
from ase.optimize import BFGS
from ase.vibrations import Vibrations
from ase.vibrations.franck_condon import (
    FranckCondon,
    FranckCondonOverlap,
    FranckCondonRecursive,
)


def equal(x, y, tolerance=0, fail=True, msg=''):
    """Compare x and y."""

    if not np.isfinite(x - y).any() or (np.abs(x - y) > tolerance).any():
        msg = (msg + '%s != %s (error: |%s| > %.9g)' %
               (x, y, x - y, tolerance))
        if fail:
            raise AssertionError(msg)
        else:
            sys.stderr.write('WARNING: %s\n' % msg)


@pytest.mark.optimize()
def test_franck_condon(testdir):
    # FCOverlap

    fco = FranckCondonOverlap()
    fcr = FranckCondonRecursive()

    # check factorial
    assert fco.factorial(8) == factorial(8)
    # the second test is useful according to the implementation
    assert fco.factorial(5) == factorial(5)
    assert fco.factorial.inv(5) == 1. / factorial(5)

    # check T=0 and n=0 equality
    S = np.array([1, 2.1, 34])
    m = 5
    assert ((fco.directT0(m, S) - fco.direct(0, m, S)) / fco.directT0(m, S) <
            1e-15).all()

    # check symmetry
    S = 2
    n = 3
    assert fco.direct(n, m, S) == fco.direct(m, n, S)

    # ---------------------------
    # specials
    S = np.array([0, 1.5])
    delta = np.sqrt(2 * S)
    for m in [2, 7]:
        equal(fco.direct0mm1(m, S)**2,
              fco.direct(1, m, S) * fco.direct(m, 0, S), 1.e-17)
        equal(fco.direct0mm1(m, S), fcr.ov0mm1(m, delta), 1.e-15)
        equal(fcr.ov0mm1(m, delta),
              fcr.ov0m(m, delta) * fcr.ov1m(m, delta), 1.e-15)
        equal(fcr.ov0mm1(m, -delta), fcr.direct0mm1(m, -delta), 1.e-15)
        equal(fcr.ov0mm1(m, delta), - fcr.direct0mm1(m, -delta), 1.e-15)

        equal(fco.direct0mm2(m, S)**2,
              fco.direct(2, m, S) * fco.direct(m, 0, S), 1.e-17)
        equal(fco.direct0mm2(m, S), fcr.ov0mm2(m, delta), 1.e-15)
        equal(fcr.ov0mm2(m, delta),
              fcr.ov0m(m, delta) * fcr.ov2m(m, delta), 1.e-15)
        equal(fco.direct0mm2(m, S), fcr.direct0mm2(m, delta), 1.e-15)

        equal(fcr.direct0mm3(m, delta),
              fcr.ov0m(m, delta) * fcr.ov3m(m, delta), 1.e-15)

        equal(fcr.ov1mm2(m, delta),
              fcr.ov1m(m, delta) * fcr.ov2m(m, delta), 1.e-15)
        equal(fcr.direct1mm2(m, delta), fcr.ov1mm2(m, delta), 1.e-15)


@pytest.fixture(scope='module')
def unrelaxed():
    atoms = molecule('CH4')
    atoms.calc = EMT()
    return atoms


@pytest.fixture(scope='module')
def forces_a(unrelaxed):
    # evaluate forces in this configuration
    return unrelaxed.get_forces()


@pytest.fixture(scope='module')
def relaxed(unrelaxed):
    atoms = unrelaxed.copy()
    atoms.calc = unrelaxed.calc
    with BFGS(atoms, logfile=None) as opt:
        opt.run(fmax=0.01)
    return atoms


@pytest.fixture()
def vibname(testdir, relaxed):
    atoms = relaxed.copy()
    atoms.calc = relaxed.calc
    name = 'vib'
    vib = Vibrations(atoms, name=name)
    vib.run()
    return name


@pytest.mark.optimize()
def test_ch4_all(forces_a, relaxed, vibname):
    """Evaluate Franck-Condon overlaps in
    a molecule suddenly exposed to a different potential"""

    # FC factor for all frequencies
    fc = FranckCondon(relaxed, vibname)
    ndof = 3 * len(relaxed)

    # by symmetry only one frequency has a non-vanishing contribution
    HR_a, _f_a = fc.get_Huang_Rhys_factors(forces_a)
    assert len(HR_a) == ndof
    assert HR_a[:-1] == pytest.approx(0, abs=1e-10)
    assert HR_a[-1] == pytest.approx(0.859989171)

    FC, freq = fc.get_Franck_Condon_factors(293, forces_a)
    assert len(FC[0]) == 2 * ndof + 1
    assert len(freq[0]) == 2 * ndof + 1


@pytest.mark.optimize()
def test_ch4_minfreq(forces_a, relaxed, vibname):
    # FC factor for relevant frequencies only
    fc = FranckCondon(relaxed, vibname, minfreq=2000)
    nrel = 4

    # single excitations
    FC, freq = fc.get_Franck_Condon_factors(293, forces_a)
    assert len(FC[0]) == 2 * nrel + 1
    assert len(freq[0]) == 2 * nrel + 1

    # include double excitations
    FC, freq = fc.get_Franck_Condon_factors(293, forces_a, 2)
    assert len(FC[1]) == 2 * nrel
    # assert len(FC[2]) == 22  # XXX why? - gives 20 in oldlibs???
    for i in range(3):
        assert len(freq[i]) == len(FC[i])



"""
IO support for the qb@ll sys format.

The positions and cell dimensions are in Bohrs.

Contributed by Rafi Ullah <rraffiu@gmail.com>
"""

from re import compile

from ase.atoms import Atoms
from ase.units import Bohr

__all__ = ['read_sys', 'write_sys']


def read_sys(fileobj):
    """
    Function to read a qb@ll sys file.

    fileobj: file (object) to read from.
    """
    a1, a2, a3, b1, b2, b3, c1, c2, c3 = fileobj.readline().split()[2:11]
    cell = []
    cell.append([float(a1) * Bohr, float(a2) * Bohr, float(a3) * Bohr])
    cell.append([float(b1) * Bohr, float(b2) * Bohr, float(b3) * Bohr])
    cell.append([float(c1) * Bohr, float(c2) * Bohr, float(c3) * Bohr])
    positions = []
    symbols = []
    reg = compile(r'(\d+|\s+)')
    line = fileobj.readline()
    while 'species' in line:
        line = fileobj.readline()
    while line:
        # The units column is ignored.
        _a, symlabel, _spec, x, y, z = line.split()[0:6]
        positions.append([float(x) * Bohr, float(y) * Bohr,
                         float(z) * Bohr])
        sym = reg.split(str(symlabel))
        symbols.append(sym[0])
        line = fileobj.readline()
    atoms = Atoms(symbols=symbols, cell=cell, positions=positions)
    return atoms


def write_sys(fileobj, atoms):
    """
    Function to write a sys file.

    fileobj: file object
        File to which output is written.
    atoms: Atoms object
        Atoms object specifying the atomic configuration.
    """
    fileobj.write('set cell')
    for i in range(3):
        d = atoms.cell[i] / Bohr
        fileobj.write((' {:6f}  {:6f}  {:6f}').format(*d))
    fileobj.write('  bohr\n')

    ch_sym = atoms.get_chemical_symbols()
    atm_nm = atoms.numbers
    a_pos = atoms.positions
    an = list(set(atm_nm))

    for i, s in enumerate(set(ch_sym)):
        fileobj.write(('species {}{} {}.xml \n').format(s, an[i], s))
    for i, (S, Z, (x, y, z)) in enumerate(zip(ch_sym, atm_nm, a_pos)):
        fileobj.write(('atom {:5} {:5}  {:12.6f} {:12.6f} {:12.6f}\
        bohr\n').format(S + str(i + 1), S + str(Z), x / Bohr, y / Bohr,
                        z / Bohr))



from functools import partial

import numpy as np
import pytest

from ase import Atoms
from ase.build import bulk, molecule
from ase.dft.kpoints import monkhorst_pack
from ase.dft.wannier import (
    Wannier,
    arbitrary_s_orbitals,
    calculate_weights,
    gram_schmidt,
    init_orbitals,
    lowdin,
    md_min,
    neighbor_k_search,
    rotation_from_projection,
    scdm,
    search_for_gamma_point,
    steepest_descent,
)
from ase.dft.wannierstate import random_orthogonal_matrix
from ase.io.cube import read_cube
from ase.lattice import (
    BCC,
    BCT,
    CRECT,
    CUB,
    FCC,
    HEX,
    HEX2D,
    LINE,
    MCL,
    MCLC,
    OBL,
    ORC,
    ORCC,
    ORCF,
    ORCI,
    RECT,
    RHL,
    SQR,
    TET,
    TRI,
)
from ase.transport.tools import dagger, normalize

calc = pytest.mark.calculator
Nk = 2


@pytest.fixture()
def rng():
    return np.random.RandomState(0)


@pytest.fixture(scope='module')
def _base_calculator_gpwfile(tmp_path_factory, factories):
    """
    Generic method to cache calculator in a file on disk.
    """
    def __base_calculator_gpwfile(atoms, filename,
                                  nbands, gpts=(8, 8, 8),
                                  kpts=(Nk, Nk, Nk)):
        factories.require('gpaw')
        import gpaw
        gpw_path = tmp_path_factory.mktemp('sub') / filename
        calc = gpaw.GPAW(
            mode='fd',
            gpts=gpts,
            nbands=nbands,
            kpts={'size': kpts, 'gamma': True},
            symmetry='off',
            txt=None)
        atoms.calc = calc
        atoms.get_potential_energy()
        calc.write(gpw_path, mode='all')
        return gpw_path
    return __base_calculator_gpwfile


@pytest.fixture(scope='module')
def _h2_calculator_gpwfile(_base_calculator_gpwfile):
    atoms = molecule('H2', pbc=True)
    atoms.center(vacuum=3.)
    gpw_path = _base_calculator_gpwfile(
        atoms=atoms,
        filename='wan_h2.gpw',
        nbands=4
    )
    return gpw_path


@pytest.fixture(scope='module')
def h2_calculator(_h2_calculator_gpwfile):
    import gpaw
    return gpaw.GPAW(_h2_calculator_gpwfile, txt=None)


@pytest.fixture(scope='module')
def _si_calculator_gpwfile(_base_calculator_gpwfile):
    atoms = bulk('Si')
    gpw_path = _base_calculator_gpwfile(
        atoms=atoms,
        filename='wan_si.gpw',
        nbands=8
    )
    return gpw_path


@pytest.fixture(scope='module')
def si_calculator(_si_calculator_gpwfile):
    import gpaw
    return gpaw.GPAW(_si_calculator_gpwfile, txt=None)


@pytest.fixture(scope='module')
def _ti_calculator_gpwfile(_base_calculator_gpwfile):
    atoms = bulk('Ti', crystalstructure='hcp')
    gpw_path = _base_calculator_gpwfile(
        atoms=atoms,
        filename='wan_ti.gpw',
        nbands=None
    )
    return gpw_path


@pytest.fixture(scope='module')
def ti_calculator(_ti_calculator_gpwfile):
    import gpaw
    return gpaw.GPAW(_ti_calculator_gpwfile, txt=None)


@pytest.fixture()
def wan(rng, h2_calculator):
    def _wan(
        atoms=None,
        calc=None,
        nwannier=2,
        fixedstates=None,
        fixedenergy=None,
        initialwannier='bloch',
        functional='std',
        gpts=(8, 8, 8),
        kpts=None,
        file=None,
        rng=rng,
        full_calc=False,
        std_calc=True,
    ):
        """
        Generate a Wannier object.

        full_calc: the provided calculator has a converged calculation
        std_calc: the default H2 calculator object is used
        """
        # If the calculator, or some fundamental parameters, are provided
        # we clearly do not want a default calculator
        if calc is not None or kpts is not None or atoms is not None:
            std_calc = False
            # Default value for kpts, if we need to generate atoms/calc
            if kpts is None:
                kpts = (Nk, Nk, Nk)

        if std_calc:
            calc = h2_calculator
            full_calc = True
        elif atoms is None and not full_calc:
            pbc = (np.array(kpts) > 1).any()
            atoms = molecule('H2', pbc=pbc)
            atoms.center(vacuum=3.)

        if calc is None:
            gpaw = pytest.importorskip('gpaw')
            calc = gpaw.GPAW(
                mode='fd',
                gpts=gpts,
                nbands=nwannier,
                kpts=kpts,
                symmetry='off',
                txt=None
            )

        if not full_calc:
            atoms.calc = calc
            atoms.get_potential_energy()

        return Wannier(
            nwannier=nwannier,
            fixedstates=fixedstates,
            fixedenergy=fixedenergy,
            calc=calc,
            initialwannier=initialwannier,
            file=None,
            functional=functional,
            rng=rng,
        )
    return _wan


def bravais_lattices():
    return [CUB(1), FCC(1), BCC(1), TET(1, 2), BCT(1, 2),
            ORC(1, 2, 3), ORCF(1, 2, 3), ORCI(1, 2, 3),
            ORCC(1, 2, 3), HEX(1, 2), RHL(1, 110),
            MCL(1, 2, 3, 70), MCLC(1, 2, 3, 70),
            TRI(1, 2, 3, 60, 70, 80), OBL(1, 2, 70),
            HEX2D(1), RECT(1, 2), CRECT(1, 70), SQR(1),
            LINE(1)]


class Paraboloid:

    def __init__(self, pos=(10., 10., 10.), shift=1.):
        self.pos = np.array(pos, dtype=complex)
        self.shift = shift

    def get_gradients(self):
        return 2 * self.pos

    def step(self, dF, updaterot=True, updatecoeff=True):
        self.pos -= dF

    def get_functional_value(self):
        return np.sum(self.pos**2) + self.shift


def unitarity_error(matrix):
    return np.abs(dagger(matrix) @ matrix - np.eye(len(matrix))).max()


def orthogonality_error(matrix):
    errors = []
    for i in range(len(matrix)):
        for j in range(i + 1, len(matrix)):
            errors.append(np.abs(dagger(matrix[i]) @ matrix[j]))
    return np.max(errors)


def normalization_error(matrix):
    old_matrix = matrix.copy()
    normalize(matrix)
    return np.abs(matrix - old_matrix).max()


def test_gram_schmidt(rng):
    matrix = rng.random((4, 4))
    assert unitarity_error(matrix) > 1
    gram_schmidt(matrix)
    assert unitarity_error(matrix) < 1e-12


def test_lowdin(rng):
    matrix = rng.random((4, 4))
    assert unitarity_error(matrix) > 1
    lowdin(matrix)
    assert unitarity_error(matrix) < 1e-12


def test_random_orthogonal_matrix(rng):
    dim = 4
    matrix = random_orthogonal_matrix(dim, rng=rng, real=True)
    assert matrix.shape[0] == matrix.shape[1]
    assert unitarity_error(matrix) < 1e-12
    matrix = random_orthogonal_matrix(dim, rng=rng, real=False)
    assert matrix.shape[0] == matrix.shape[1]
    assert unitarity_error(matrix) < 1e-12


def test_neighbor_k_search():
    kpt_kc = monkhorst_pack((4, 4, 4))
    Gdir_dc = [[1, 0, 0], [0, 1, 0], [0, 0, 1],
               [1, 1, 0], [1, 0, 1], [0, 1, 1]]
    tol = 1e-4
    for d, Gdir_c in enumerate(Gdir_dc):
        for k, k_c in enumerate(kpt_kc):
            kk, k0 = neighbor_k_search(k_c, Gdir_c, kpt_kc, tol=tol)
            assert np.linalg.norm(kpt_kc[kk] - k_c - Gdir_c + k0) < tol


@pytest.mark.parametrize('lat', bravais_lattices())
def test_calculate_weights(lat):
    # Equation from Berghold et al. PRB v61 n15 (2000)
    tol = 1e-5
    cell = lat.tocell()
    g = cell @ cell.T
    w, G = calculate_weights(cell, normalize=False)

    errors = []
    for i in range(3):
        for j in range(3):
            errors.append(np.abs((w * G[:, i] @ G[:, j]) - g[i, j]))

    assert np.max(errors) < tol


def test_steepest_descent():
    tol = 1e-6
    step = 0.1
    func = Paraboloid(pos=np.array([10, 10, 10], dtype=float), shift=1.)
    steepest_descent(func=func, step=step, tolerance=tol)
    assert func.get_functional_value() == pytest.approx(1, abs=1e-5)


def test_md_min():
    tol = 1e-8
    step = 0.1
    func = Paraboloid(pos=np.array([10, 10, 10], dtype=complex), shift=1.)
    md_min(func=func, step=step, tolerance=tol,
           max_iter=1e6)
    assert func.get_functional_value() == pytest.approx(1, abs=1e-5)


def test_rotation_from_projection(rng):
    proj_nw = rng.random((6, 4))
    assert unitarity_error(proj_nw[:int(min(proj_nw.shape))]) > 1
    U_ww, C_ul = rotation_from_projection(proj_nw, fixed=2, ortho=True)
    assert unitarity_error(U_ww) < 1e-10, 'U_ww not unitary'
    assert orthogonality_error(C_ul.T) < 1e-10, 'C_ul columns not orthogonal'
    assert normalization_error(C_ul) < 1e-10, 'C_ul not normalized'
    U_ww, C_ul = rotation_from_projection(proj_nw, fixed=2, ortho=False)
    assert normalization_error(U_ww) < 1e-10, 'U_ww not normalized'


@pytest.mark.calculator_lite()
def test_save(tmpdir, wan):
    wanf = wan(nwannier=4, fixedstates=2, initialwannier='bloch')
    jsonfile = tmpdir.join('wanf.json')
    f1 = wanf.get_functional_value()
    wanf.save(jsonfile)
    wanf.initialize(file=jsonfile, initialwannier='bloch')
    assert pytest.approx(f1) == wanf.get_functional_value()


@pytest.mark.parametrize('lat', bravais_lattices())
def test_get_radii(lat, wan):
    # Sanity check, the Wannier functions' spread should always be positive.
    # Also, make sure that the method does not fail for any lattice.
    if ((lat.tocell() == FCC(a=1).tocell()).all() or
            (lat.tocell() == ORCF(a=1, b=2, c=3).tocell()).all()):
        pytest.skip("Lattices not supported by this function,"
                    " use get_spreads() instead.")
    atoms = molecule('H2', pbc=True)
    atoms.cell = lat.tocell()
    atoms.center(vacuum=3.)
    wanf = wan(nwannier=4, fixedstates=2, atoms=atoms, initialwannier='bloch')
    assert all(wanf.get_radii() > 0)


@pytest.mark.parametrize('lat', bravais_lattices())
def test_get_spreads(lat, wan):
    # Sanity check, the Wannier functions' spread should always be positive.
    # Also, make sure that the method does not fail for any lattice.
    atoms = molecule('H2', pbc=True)
    atoms.cell = lat.tocell()
    atoms.center(vacuum=3.)
    wanf = wan(nwannier=4, fixedstates=2, atoms=atoms, initialwannier='bloch')
    assert all(wanf.get_spreads() > 0)


@pytest.mark.parametrize('fun', ['std', 'var'])
def test_get_functional_value(fun, wan):
    # Only testing if the functional scales with the number of functions
    wan1 = wan(nwannier=3, functional=fun)
    f1 = wan1.get_functional_value()
    wan2 = wan(nwannier=4)
    f2 = wan2.get_functional_value()
    assert f1 < f2


@pytest.mark.calculator_lite()
@calc('gpaw')
def test_get_centers(factory):
    # Rough test on the position of the Wannier functions' centers
    gpaw = pytest.importorskip('gpaw')
    calc = gpaw.GPAW(mode='fd', gpts=(32, 32, 32), nbands=4, txt=None)
    atoms = molecule('H2', calculator=calc)
    atoms.center(vacuum=3.)
    atoms.get_potential_energy()
    wanf = Wannier(nwannier=2, calc=calc, initialwannier='bloch')
    centers = wanf.get_centers()
    com = atoms.get_center_of_mass()
    assert np.abs(centers - [com, com]).max() < 1e-4


def test_write_cube_default(wan, h2_calculator, testdir):
    # Chek the value saved in the CUBE file and the atoms object.
    # The default saved value is the absolute value of the Wannier function,
    # and the supercell is repeated per the number of k-points in each
    # direction.
    atoms = h2_calculator.atoms
    wanf = wan(calc=h2_calculator, full_calc=True)
    index = 0

    # It returns some errors when using file objects, so we use a string
    cubefilename = 'wanf.cube'
    wanf.write_cube(index, cubefilename)
    with open(cubefilename) as inputfile:
        content = read_cube(inputfile)
    assert pytest.approx(content['atoms'].cell.array) == atoms.cell.array * 2
    assert pytest.approx(content['data']) == abs(wanf.get_function(index))


def test_write_cube_angle(wan, testdir):
    # Check that the complex phase is correctly saved to the CUBE file, together
    # with the right atoms object.
    atoms = molecule('H2')
    atoms.center(vacuum=3.)
    wanf = wan(atoms=atoms, kpts=(1, 1, 1))
    index = 0

    # It returns some errors when using file objects, so we use a string
    cubefilename = 'wanf.cube'
    wanf.write_cube(index, cubefilename, angle=True)
    with open(cubefilename) as inputfile:
        content = read_cube(inputfile)
    assert pytest.approx(content['atoms'].cell.array) == atoms.cell.array
    assert pytest.approx(content['data']) == np.angle(wanf.get_function(index))


def test_write_cube_repeat(wan, testdir):
    # Check that the repeated supercell and Wannier functions are correctly
    # saved to the CUBE file, together with the right atoms object.
    atoms = molecule('H2')
    atoms.center(vacuum=3.)
    wanf = wan(atoms=atoms, kpts=(1, 1, 1))
    index = 0
    repetition = [4, 4, 4]

    # It returns some errors when using file objects, so we use simple filename
    cubefilename = 'wanf.cube'
    wanf.write_cube(index, cubefilename, repeat=repetition)

    with open(cubefilename) as inputfile:
        content = read_cube(inputfile)
    assert pytest.approx(content['atoms'].cell.array) == \
        (atoms * repetition).cell.array
    assert pytest.approx(content['data']) == \
        abs(wanf.get_function(index, repetition))


def test_localize(wan):
    wanf = wan(initialwannier='random')
    fvalue = wanf.get_functional_value()
    wanf.localize()
    assert wanf.get_functional_value() > fvalue


def test_get_spectral_weight_bloch(wan):
    nwannier = 4
    wanf = wan(initialwannier='bloch', nwannier=nwannier)
    for i in range(nwannier):
        assert wanf.get_spectral_weight(i)[:, i].sum() == pytest.approx(1)


def test_get_spectral_weight_random(wan, rng):
    nwannier = 4
    wanf = wan(initialwannier='random', nwannier=nwannier, rng=rng)
    for i in range(nwannier):
        assert wanf.get_spectral_weight(i).sum() == pytest.approx(1)


@pytest.mark.calculator_lite()
def test_get_pdos(wan):
    nwannier = 4
    gpaw = pytest.importorskip('gpaw')
    calc = gpaw.GPAW(mode='fd', gpts=(16, 16, 16), nbands=nwannier, txt=None)
    atoms = molecule('H2')
    atoms.center(vacuum=3.)
    atoms.calc = calc
    atoms.get_potential_energy()
    wanf = wan(atoms=atoms, calc=calc,
               nwannier=nwannier, initialwannier='bloch')
    eig_n = calc.get_eigenvalues()
    for i in range(nwannier):
        pdos_n = wanf.get_pdos(w=i, energies=eig_n, width=0.001)
        assert pdos_n[i] != pytest.approx(0)


def test_translate(wan, h2_calculator):
    nwannier = 2
    calc = h2_calculator
    atoms = calc.get_atoms()
    wanf = wan(nwannier=nwannier, initialwannier='bloch',
               calc=calc, full_calc=True)
    wanf.translate_all_to_cell(cell=[0, 0, 0])
    c0_w = wanf.get_centers()
    for i in range(nwannier):
        c2_w = np.delete(wanf.get_centers(), i, 0)
        wanf.translate(w=i, R=[1, 1, 1])
        c1_w = wanf.get_centers()
        assert np.linalg.norm(c1_w[i] - c0_w[i]) == \
            pytest.approx(np.linalg.norm(atoms.cell.array.diagonal()))
        c1_w = np.delete(c1_w, i, 0)
        assert c1_w == pytest.approx(c2_w)


def test_translate_to_cell(wan, h2_calculator):
    nwannier = 2
    calc = h2_calculator
    atoms = calc.get_atoms()
    wanf = wan(nwannier=nwannier, initialwannier='bloch',
               calc=calc, full_calc=True)
    for i in range(nwannier):
        wanf.translate_to_cell(w=i, cell=[0, 0, 0])
        c0_w = wanf.get_centers()
        assert (c0_w[i] < atoms.cell.array.diagonal()).all()
        wanf.translate_to_cell(w=i, cell=[1, 1, 1])
        c1_w = wanf.get_centers()
        assert (c1_w[i] > atoms.cell.array.diagonal()).all()
        assert np.linalg.norm(c1_w[i] - c0_w[i]) == \
            pytest.approx(np.linalg.norm(atoms.cell.array.diagonal()))
        c0_w = np.delete(c0_w, i, 0)
        c1_w = np.delete(c1_w, i, 0)
        assert c0_w == pytest.approx(c1_w)


def test_translate_all_to_cell(wan, h2_calculator):
    nwannier = 2
    calc = h2_calculator
    atoms = calc.get_atoms()
    wanf = wan(nwannier=nwannier, initialwannier='bloch',
               calc=calc, full_calc=True)
    wanf.translate_all_to_cell(cell=[0, 0, 0])
    c0_w = wanf.get_centers()
    assert (c0_w < atoms.cell.array.diagonal()).all()
    wanf.translate_all_to_cell(cell=[1, 1, 1])
    c1_w = wanf.get_centers()
    assert (c1_w > atoms.cell.array.diagonal()).all()
    for i in range(nwannier):
        assert np.linalg.norm(c1_w[i] - c0_w[i]) == \
            pytest.approx(np.linalg.norm(atoms.cell.array.diagonal()))


def test_distances(wan, h2_calculator):
    nwannier = 2
    calc = h2_calculator
    atoms = calc.get_atoms()
    wanf = wan(nwannier=nwannier, initialwannier='bloch')
    cent_w = wanf.get_centers()
    dist_ww = wanf.distances([0, 0, 0])
    dist1_ww = wanf.distances([1, 1, 1])
    for i in range(nwannier):
        assert dist_ww[i, i] == pytest.approx(0)
        assert dist1_ww[i, i] == pytest.approx(np.linalg.norm(atoms.cell.array))
        for j in range(i + 1, nwannier):
            assert dist_ww[i, j] == dist_ww[j, i]
            assert dist_ww[i, j] == \
                pytest.approx(np.linalg.norm(cent_w[i] - cent_w[j]))


def test_get_hopping_bloch(wan):
    nwannier = 4
    wanf = wan(nwannier=nwannier, initialwannier='bloch')
    hop0_ww = wanf.get_hopping([0, 0, 0])
    hop1_ww = wanf.get_hopping([1, 1, 1])
    for i in range(nwannier):
        assert hop0_ww[i, i] != 0
        assert hop1_ww[i, i] != 0
        for j in range(i + 1, nwannier):
            assert hop0_ww[i, j] == 0
            assert hop1_ww[i, j] == 0
            assert hop0_ww[i, j] == hop0_ww[j, i]
            assert hop1_ww[i, j] == hop1_ww[j, i]


def test_get_hopping_random(wan, rng):
    nwannier = 4
    wanf = wan(nwannier=nwannier, initialwannier='random')
    hop0_ww = wanf.get_hopping([0, 0, 0])
    hop1_ww = wanf.get_hopping([1, 1, 1])
    for i in range(nwannier):
        for j in range(i + 1, nwannier):
            assert np.abs(hop0_ww[i, j]) == pytest.approx(np.abs(hop0_ww[j, i]))
            assert np.abs(hop1_ww[i, j]) == pytest.approx(np.abs(hop1_ww[j, i]))


def test_get_hamiltonian_bloch(wan):
    nwannier = 4
    atoms = molecule('H2', pbc=True)
    atoms.center(vacuum=3.)
    kpts = (2, 2, 2)
    number_kpts = kpts[0] * kpts[1] * kpts[2]
    wanf = wan(atoms=atoms, kpts=kpts,
               nwannier=nwannier, initialwannier='bloch')
    for k in range(number_kpts):
        H_ww = wanf.get_hamiltonian(k=k)
        for i in range(nwannier):
            assert H_ww[i, i] != 0
            for j in range(i + 1, nwannier):
                assert H_ww[i, j] == 0
                assert H_ww[i, j] == pytest.approx(H_ww[j, i])


def test_get_hamiltonian_random(wan, rng):
    nwannier = 4
    atoms = molecule('H2', pbc=True)
    atoms.center(vacuum=3.)
    kpts = (2, 2, 2)
    number_kpts = kpts[0] * kpts[1] * kpts[2]
    wanf = wan(atoms=atoms, kpts=kpts, rng=rng,
               nwannier=nwannier, initialwannier='random')
    for k in range(number_kpts):
        H_ww = wanf.get_hamiltonian(k=k)
        for i in range(nwannier):
            for j in range(i + 1, nwannier):
                assert np.abs(H_ww[i, j]) == pytest.approx(np.abs(H_ww[j, i]))


def test_get_hamiltonian_kpoint(wan, rng, h2_calculator):
    nwannier = 4
    calc = h2_calculator
    atoms = calc.get_atoms()
    wanf = wan(nwannier=nwannier, initialwannier='random')
    kpts = atoms.cell.bandpath(density=50).cartesian_kpts()
    for kpt_c in kpts:
        H_ww = wanf.get_hamiltonian_kpoint(kpt_c=kpt_c)
        for i in range(nwannier):
            for j in range(i + 1, nwannier):
                assert np.abs(H_ww[i, j]) == pytest.approx(np.abs(H_ww[j, i]))


def test_get_function(wan):
    nwannier = 2
    gpts_np = np.array((8, 8, 8))
    wanf = wan(nwannier=nwannier, initialwannier='bloch')
    assert (wanf.get_function(index=[0, 0]) == 0).all()
    assert wanf.get_function(index=[0, 1]) + wanf.get_function(index=[1, 0]) \
        == pytest.approx(wanf.get_function(index=[1, 1]))
    for i in range(nwannier):
        assert (gpts_np * Nk == wanf.get_function(index=i).shape).all()
        assert (gpts_np * [1, 2, 3] ==
                wanf.get_function(index=i, repeat=[1, 2, 3]).shape).all()


# Make sure this can run in CI: https://gitlab.com/ase/ase/-/issues/1100
# @pytest.mark.calculator_lite
@pytest.mark.parametrize('fun', ['std', 'var'])
def test_get_gradients(fun, wan, rng):
    wanf = wan(nwannier=4, fixedstates=2, gpts=(12, 12, 12), kpts=(1, 1, 1),
               initialwannier='bloch', functional=fun)
    # create an anti-hermitian array/matrix
    step = rng.random(wanf.get_gradients().size) + \
        1.j * rng.random(wanf.get_gradients().size)
    step *= 1e-8
    step -= dagger(step)
    f1 = wanf.get_functional_value()
    wanf.step(step)
    f2 = wanf.get_functional_value()
    assert (np.abs((f2 - f1) / step).ravel() -
            np.abs(wanf.get_gradients())).max() < 1e-4


@pytest.mark.parametrize('init', ['bloch', 'random', 'orbitals', 'scdm'])
def test_initialwannier(init, wan, ti_calculator):
    # dummy check to run the module with different initialwannier methods
    wanf = wan(calc=ti_calculator, full_calc=True,
               initialwannier=init, nwannier=14, fixedstates=12)
    assert wanf.get_functional_value() > 0


def test_nwannier_auto(wan, ti_calculator):
    """ Test 'auto' value for parameter 'nwannier'. """
    partial_wan = partial(
        wan,
        calc=ti_calculator,
        full_calc=True,
        initialwannier='bloch',
        nwannier='auto'
    )

    # Check default value
    wanf = partial_wan()
    assert wanf.nwannier == 15

    # Check value setting fixedenergy
    wanf = partial_wan(fixedenergy=0)
    assert wanf.nwannier == 15
    wanf = partial_wan(fixedenergy=5)
    assert wanf.nwannier == 18

    # Check value setting fixedstates
    number_kpts = Nk**3
    list_fixedstates = [14] * number_kpts
    list_fixedstates[Nk] = 18
    wanf = partial_wan(fixedstates=list_fixedstates)
    assert wanf.nwannier == 18


def test_arbitrary_s_orbitals(rng):
    atoms = Atoms('3H', positions=[[0, 0, 0],
                                   [1, 1.5, 1],
                                   [2, 3, 0]])
    orbs = arbitrary_s_orbitals(atoms, 10, rng)

    atoms.append('H')
    s_pos = atoms.get_scaled_positions()
    for orb in orbs:
        # Test if they are actually s-orbitals
        assert orb[1] == 0

        # Read random position
        x, y, z = orb[0]
        s_pos[-1] = [x, y, z]
        atoms.set_scaled_positions(s_pos)

        # Use dummy H atom to measure distance from any other atom
        dists = atoms.get_distances(
            a=-1,
            indices=range(atoms.get_global_number_of_atoms() - 1))

        # Test that the s-orbital is close to at least one atom
        assert (dists < 1.5).any()


def test_init_orbitals_h2(rng):
    # Check that the initial orbitals for H2 are as many as requested and they
    # are all s-orbitals (l=0).
    atoms = molecule('H2')
    atoms.center(vacuum=3.)
    ntot = 2
    orbs = init_orbitals(atoms=atoms, ntot=ntot, rng=rng)
    angular_momenta = [orb[1] for orb in orbs]
    assert sum(l_ * 2 + 1 for l_ in angular_momenta) == ntot
    assert angular_momenta == [0] * ntot


def test_init_orbitals_ti(rng):
    # Check that the initial orbitals for Ti bulk are as many as requested and
    # there are both s-orbitals (l=0) and d-orbitals (l=2).
    atoms = bulk('Ti')
    ntot = 14
    orbs = init_orbitals(atoms=atoms, ntot=ntot, rng=rng)
    angular_momenta = [orb[1] for orb in orbs]
    assert sum(l_ * 2 + 1 for l_ in angular_momenta) == ntot
    assert 0 in angular_momenta
    assert 2 in angular_momenta


def test_search_for_gamma_point():
    list_with_gamma = [[-1.0, -1.0, -1.0],
                       [0.0, 0.0, 0.0],
                       [0.1, 0.0, 0.0],
                       [1.5, 2.5, 0.5]]
    gamma_idx = search_for_gamma_point(list_with_gamma)
    assert gamma_idx == 1

    list_without_gamma = [[-1.0, -1.0, -1.0],
                          [0.1, 0.0, 0.0],
                          [1.5, 2.5, 0.5]]
    gamma_idx = search_for_gamma_point(list_without_gamma)
    assert gamma_idx is None


def test_scdm(ti_calculator):
    calc = ti_calculator
    Nw = 14
    ps = calc.get_pseudo_wave_function(band=Nw, kpt=0, spin=0)
    Ng = ps.size
    kpt_kc = calc.get_bz_k_points()
    number_kpts = len(kpt_kc)
    nbands = calc.get_number_of_bands()
    pseudo_nkG = np.zeros((nbands, number_kpts, Ng), dtype=np.complex128)
    for k in range(number_kpts):
        for n in range(nbands):
            pseudo_nkG[n, k] = calc.get_pseudo_wave_function(
                band=n, kpt=k, spin=0).ravel()
    fixed_k = [Nw - 2] * number_kpts
    C_kul, U_kww = scdm(pseudo_nkG, kpts=kpt_kc,
                        fixed_k=fixed_k, Nw=Nw)
    for k in range(number_kpts):
        assert unitarity_error(U_kww[k]) < 1e-10, 'U_ww not unitary'
        assert orthogonality_error(C_kul[k].T) < 1e-10, \
            'C_ul columns not orthogonal'
        assert normalization_error(C_kul[k]) < 1e-10, 'C_ul not normalized'


@pytest.mark.xfail()
def test_get_optimal_nwannier(wan, si_calculator):
    """ Test method to compute the optimal 'nwannier' value. """

    wanf = wan(calc=si_calculator, full_calc=True,
               initialwannier='bloch', nwannier='auto', fixedenergy=1)

    # Test with default parameters
    opt_nw = wanf.get_optimal_nwannier()
    assert opt_nw == 7

    # Test with non-default parameters.
    # This is mostly to test that is does actually support this parameters,
    # it's not really testing the actual result.
    opt_nw = wanf.get_optimal_nwannier(nwrange=10)
    assert opt_nw == 7
    opt_nw = wanf.get_optimal_nwannier(tolerance=1e-2)
    assert opt_nw == 8

    # This should give same result since the initialwannier does not include
    # randomness.
    opt_nw = wanf.get_optimal_nwannier(random_reps=10)
    assert opt_nw == 7

    # Test with random repetitions, just test if it runs.
    wanf = wan(calc=si_calculator, full_calc=True,
               initialwannier='orbitals', nwannier='auto', fixedenergy=0)
    opt_nw = wanf.get_optimal_nwannier(random_reps=10)
    assert opt_nw >= 0


@pytest.mark.xfail()
def test_spread_contributions(wan):
    # Only a test on a constant value to make sure it does not deviate too much
    wan1 = wan()
    test_values_w = wan1._spread_contributions()
    ref_values_w = [2.28535569, 0.04660427]
    assert test_values_w == pytest.approx(ref_values_w, abs=1e-4)
